# coding: utf8
from __future__ import unicode_literals

from ...symbols import LEMMA, PRON_LEMMA

# Source: Danish Universal Dependencies and http://fjern-uv.dk/pronom.php

# Note: The Danish Universal Dependencies specify Case=Acc for all instances
# of "den"/"det" even when the case is in fact "Nom". In the rules below, Case
# is left unspecified for "den" and "det".

MORPH_RULES = {
    "PRON": {
        "jeg": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "One",
            "Number": "Sing",
            "Case": "Nom",
            "Gender": "Com",
        },  # Case=Nom|Gender=Com|Number=Sing|Person=1|PronType=Prs
        "mig": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "One",
            "Number": "Sing",
            "Case": "Acc",
            "Gender": "Com",
        },  # Case=Acc|Gender=Com|Number=Sing|Person=1|PronType=Prs
        "min": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "One",
            "Number": "Sing",
            "Poss": "Yes",
            "Gender": "Com",
        },  # Gender=Com|Number=Sing|Number[psor]=Sing|Person=1|Poss=Yes|PronType=Prs
        "mit": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "One",
            "Number": "Sing",
            "Poss": "Yes",
            "Gender": "Neut",
        },  # Gender=Neut|Number=Sing|Number[psor]=Sing|Person=1|Poss=Yes|PronType=Prs
        "vor": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "One",
            "Number": "Sing",
            "Poss": "Yes",
            "Gender": "Com",
        },  # Gender=Com|Number=Sing|Number[psor]=Plur|Person=1|Poss=Yes|PronType=Prs|Style=Form
        "vort": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "One",
            "Number": "Sing",
            "Poss": "Yes",
            "Gender": "Neut",
        },  # Gender=Neut|Number=Sing|Number[psor]=Plur|Person=1|Poss=Yes|PronType=Prs|Style=Form
        "du": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Two",
            "Number": "Sing",
            "Case": "Nom",
            "Gender": "Com",
        },  # Case=Nom|Gender=Com|Number=Sing|Person=2|PronType=Prs
        "dig": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Two",
            "Number": "Sing",
            "Case": "Acc",
            "Gender": "Com",
        },  # Case=Acc|Gender=Com|Number=Sing|Person=2|PronType=Prs
        "din": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Two",
            "Number": "Sing",
            "Poss": "Yes",
            "Gender": "Com",
        },  # Gender=Com|Number=Sing|Number[psor]=Sing|Person=2|Poss=Yes|PronType=Prs
        "dit": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Two",
            "Number": "Sing",
            "Poss": "Yes",
            "Gender": "Neut",
        },  # Gender=Neut|Number=Sing|Number[psor]=Sing|Person=2|Poss=Yes|PronType=Prs
        "han": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Number": "Sing",
            "Case": "Nom",
            "Gender": "Com",
        },  # Case=Nom|Gender=Com|Number=Sing|Person=3|PronType=Prs
        "hun": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Number": "Sing",
            "Case": "Nom",
            "Gender": "Com",
        },  # Case=Nom|Gender=Com|Number=Sing|Person=3|PronType=Prs
        "den": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Number": "Sing",
            "Gender": "Com",
        },  # Case=Acc|Gender=Com|Number=Sing|Person=3|PronType=Prs, See note above.
        "det": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Number": "Sing",
            "Gender": "Neut",
        },  # Case=Acc|Gender=Neut|Number=Sing|Person=3|PronType=Prs See note above.
        "ham": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Number": "Sing",
            "Case": "Acc",
            "Gender": "Com",
        },  # Case=Acc|Gender=Com|Number=Sing|Person=3|PronType=Prs
        "hende": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Number": "Sing",
            "Case": "Acc",
            "Gender": "Com",
        },  # Case=Acc|Gender=Com|Number=Sing|Person=3|PronType=Prs
        "sin": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Number": "Sing",
            "Poss": "Yes",
            "Gender": "Com",
            "Reflex": "Yes",
        },  # Gender=Com|Number=Sing|Number[psor]=Sing|Person=3|Poss=Yes|PronType=Prs|Reflex=Yes
        "sit": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Number": "Sing",
            "Poss": "Yes",
            "Gender": "Neut",
            "Reflex": "Yes",
        },  # Gender=Neut|Number=Sing|Number[psor]=Sing|Person=3|Poss=Yes|PronType=Prs|Reflex=Yes
        "vi": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "One",
            "Number": "Plur",
            "Case": "Nom",
            "Gender": "Com",
        },  # Case=Nom|Gender=Com|Number=Plur|Person=1|PronType=Prs
        "os": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "One",
            "Number": "Plur",
            "Case": "Acc",
            "Gender": "Com",
        },  # Case=Acc|Gender=Com|Number=Plur|Person=1|PronType=Prs
        "mine": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "One",
            "Number": "Plur",
            "Poss": "Yes",
        },  # Number=Plur|Number[psor]=Sing|Person=1|Poss=Yes|PronType=Prs
        "vore": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "One",
            "Number": "Plur",
            "Poss": "Yes",
        },  # Number=Plur|Number[psor]=Plur|Person=1|Poss=Yes|PronType=Prs|Style=Form
        "I": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Two",
            "Number": "Plur",
            "Case": "Nom",
            "Gender": "Com",
        },  # Case=Nom|Gender=Com|Number=Plur|Person=2|PronType=Prs
        "jer": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Two",
            "Number": "Plur",
            "Case": "Acc",
            "Gender": "Com",
        },  # Case=Acc|Gender=Com|Number=Plur|Person=2|PronType=Prs
        "dine": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Two",
            "Number": "Plur",
            "Poss": "Yes",
        },  # Number=Plur|Number[psor]=Sing|Person=2|Poss=Yes|PronType=Prs
        "de": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Number": "Plur",
            "Case": "Nom",
        },  # Case=Nom|Number=Plur|Person=3|PronType=Prs
        "dem": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Number": "Plur",
            "Case": "Acc",
        },  # Case=Acc|Number=Plur|Person=3|PronType=Prs
        "sine": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Number": "Plur",
            "Poss": "Yes",
            "Reflex": "Yes",
        },  # Number=Plur|Number[psor]=Sing|Person=3|Poss=Yes|PronType=Prs|Reflex=Yes
        "vores": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "One",
            "Poss": "Yes",
        },  # Number[psor]=Plur|Person=1|Poss=Yes|PronType=Prs
        "De": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Two",
            "Case": "Nom",
            "Gender": "Com",
        },  # Case=Nom|Gender=Com|Person=2|Polite=Form|PronType=Prs
        "Dem": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Two",
            "Case": "Acc",
            "Gender": "Com",
        },  # Case=Acc|Gender=Com|Person=2|Polite=Form|PronType=Prs
        "Deres": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Two",
            "Poss": "Yes",
        },  # Person=2|Polite=Form|Poss=Yes|PronType=Prs
        "jeres": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Two",
            "Poss": "Yes",
        },  # Number[psor]=Plur|Person=2|Poss=Yes|PronType=Prs
        "sig": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Case": "Acc",
            "Reflex": "Yes",
        },  # Case=Acc|Person=3|PronType=Prs|Reflex=Yes
        "hans": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Poss": "Yes",
        },  # Number[psor]=Sing|Person=3|Poss=Yes|PronType=Prs
        "hendes": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Poss": "Yes",
        },  # Number[psor]=Sing|Person=3|Poss=Yes|PronType=Prs
        "dens": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Poss": "Yes",
        },  # Number[psor]=Sing|Person=3|Poss=Yes|PronType=Prs
        "dets": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Poss": "Yes",
        },  # Number[psor]=Sing|Person=3|Poss=Yes|PronType=Prs
        "deres": {
            LEMMA: PRON_LEMMA,
            "PronType": "Prs",
            "Person": "Three",
            "Poss": "Yes",
        },  # Number[psor]=Plur|Person=3|Poss=Yes|PronType=Prs
    },
    "VERB": {
        "er": {LEMMA: "være", "VerbForm": "Fin", "Tense": "Pres"},
        "var": {LEMMA: "være", "VerbForm": "Fin", "Tense": "Past"},
    },
}

for tag, rules in MORPH_RULES.items():
    for key, attrs in dict(rules).items():
        rules[key.title()] = attrs



# coding: utf-8
from __future__ import unicode_literals

import pytest


@pytest.mark.parametrize("text", ["lorem ipsum"])
def test_tokenizer_splits_single_space(tokenizer, text):
    tokens = tokenizer(text)
    assert len(tokens) == 2


@pytest.mark.parametrize("text", ["lorem  ipsum"])
def test_tokenizer_splits_double_space(tokenizer, text):
    tokens = tokenizer(text)
    assert len(tokens) == 3
    assert tokens[1].text == " "


@pytest.mark.parametrize("text", ["lorem ipsum  "])
def test_tokenizer_handles_double_trailing_ws(tokenizer, text):
    tokens = tokenizer(text)
    assert repr(tokens.text_with_ws) == repr(text)


@pytest.mark.parametrize("text", ["lorem\nipsum"])
def test_tokenizer_splits_newline(tokenizer, text):
    tokens = tokenizer(text)
    assert len(tokens) == 3
    assert tokens[1].text == "\n"


@pytest.mark.parametrize("text", ["lorem \nipsum"])
def test_tokenizer_splits_newline_space(tokenizer, text):
    tokens = tokenizer(text)
    assert len(tokens) == 3


@pytest.mark.parametrize("text", ["lorem  \nipsum"])
def test_tokenizer_splits_newline_double_space(tokenizer, text):
    tokens = tokenizer(text)
    assert len(tokens) == 3


@pytest.mark.parametrize("text", ["lorem \n ipsum"])
def test_tokenizer_splits_newline_space_wrap(tokenizer, text):
    tokens = tokenizer(text)
    assert len(tokens) == 3



# coding: utf-8
from __future__ import unicode_literals

import pytest
import numpy
from spacy.attrs import IS_ALPHA, IS_DIGIT, IS_LOWER, IS_PUNCT, IS_TITLE, IS_STOP
from spacy.symbols import VERB
from spacy.vocab import Vocab
from spacy.tokens import Doc

from ..util import get_doc


@pytest.fixture
def doc(en_tokenizer):
    # fmt: off
    text = "This is a sentence. This is another sentence. And a third."
    heads = [1, 0, 1, -2, -3, 1, 0, 1, -2, -3, 0, 1, -2, -1]
    deps = ["nsubj", "ROOT", "det", "attr", "punct", "nsubj", "ROOT", "det",
            "attr", "punct", "ROOT", "det", "npadvmod", "punct"]
    # fmt: on
    tokens = en_tokenizer(text)
    return get_doc(tokens.vocab, words=[t.text for t in tokens], heads=heads, deps=deps)


def test_doc_token_api_strings(en_tokenizer):
    text = "Give it back! He pleaded."
    pos = ["VERB", "PRON", "PART", "PUNCT", "PRON", "VERB", "PUNCT"]
    heads = [0, -1, -2, -3, 1, 0, -1]
    deps = ["ROOT", "dobj", "prt", "punct", "nsubj", "ROOT", "punct"]

    tokens = en_tokenizer(text)
    doc = get_doc(
        tokens.vocab, words=[t.text for t in tokens], pos=pos, heads=heads, deps=deps
    )
    assert doc[0].orth_ == "Give"
    assert doc[0].text == "Give"
    assert doc[0].text_with_ws == "Give "
    assert doc[0].lower_ == "give"
    assert doc[0].shape_ == "Xxxx"
    assert doc[0].prefix_ == "G"
    assert doc[0].suffix_ == "ive"
    assert doc[0].pos_ == "VERB"
    assert doc[0].dep_ == "ROOT"


def test_doc_token_api_flags(en_tokenizer):
    text = "Give it back! He pleaded."
    tokens = en_tokenizer(text)
    assert tokens[0].check_flag(IS_ALPHA)
    assert not tokens[0].check_flag(IS_DIGIT)
    assert tokens[0].check_flag(IS_TITLE)
    assert tokens[1].check_flag(IS_LOWER)
    assert tokens[3].check_flag(IS_PUNCT)
    assert tokens[2].check_flag(IS_STOP)
    assert not tokens[5].check_flag(IS_STOP)
    # TODO: Test more of these, esp. if a bug is found


@pytest.mark.parametrize("text", ["Give it back! He pleaded."])
def test_doc_token_api_prob_inherited_from_vocab(en_tokenizer, text):
    word = text.split()[0]
    en_tokenizer.vocab[word].prob = -1
    tokens = en_tokenizer(text)
    assert tokens[0].prob != 0


@pytest.mark.parametrize("text", ["one two"])
def test_doc_token_api_str_builtin(en_tokenizer, text):
    tokens = en_tokenizer(text)
    assert str(tokens[0]) == text.split(" ")[0]
    assert str(tokens[1]) == text.split(" ")[1]


def test_doc_token_api_is_properties(en_vocab):
    doc = Doc(en_vocab, words=["Hi", ",", "my", "email", "is", "test@me.com"])
    assert doc[0].is_title
    assert doc[0].is_alpha
    assert not doc[0].is_digit
    assert doc[1].is_punct
    assert doc[3].is_ascii
    assert not doc[3].like_url
    assert doc[4].is_lower
    assert doc[5].like_email


def test_doc_token_api_vectors():
    vocab = Vocab()
    vocab.reset_vectors(width=2)
    vocab.set_vector("apples", vector=numpy.asarray([0.0, 2.0], dtype="f"))
    vocab.set_vector("oranges", vector=numpy.asarray([0.0, 1.0], dtype="f"))
    doc = Doc(vocab, words=["apples", "oranges", "oov"])
    assert doc.has_vector
    assert doc[0].has_vector
    assert doc[1].has_vector
    assert not doc[2].has_vector
    apples_norm = (0 * 0 + 2 * 2) ** 0.5
    oranges_norm = (0 * 0 + 1 * 1) ** 0.5
    cosine = ((0 * 0) + (2 * 1)) / (apples_norm * oranges_norm)
    assert doc[0].similarity(doc[1]) == cosine


def test_doc_token_api_ancestors(en_tokenizer):
    # the structure of this sentence depends on the English annotation scheme
    text = "Yesterday I saw a dog that barked loudly."
    heads = [2, 1, 0, 1, -2, 1, -2, -1, -6]
    tokens = en_tokenizer(text)
    doc = get_doc(tokens.vocab, words=[t.text for t in tokens], heads=heads)
    assert [t.text for t in doc[6].ancestors] == ["dog", "saw"]
    assert [t.text for t in doc[1].ancestors] == ["saw"]
    assert [t.text for t in doc[2].ancestors] == []

    assert doc[2].is_ancestor(doc[7])
    assert not doc[6].is_ancestor(doc[2])


def test_doc_token_api_head_setter(en_tokenizer):
    # the structure of this sentence depends on the English annotation scheme
    text = "Yesterday I saw a dog that barked loudly."
    heads = [2, 1, 0, 1, -2, 1, -2, -1, -6]
    tokens = en_tokenizer(text)
    doc = get_doc(tokens.vocab, words=[t.text for t in tokens], heads=heads)

    assert doc[6].n_lefts == 1
    assert doc[6].n_rights == 1
    assert doc[6].left_edge.i == 5
    assert doc[6].right_edge.i == 7

    assert doc[4].n_lefts == 1
    assert doc[4].n_rights == 1
    assert doc[4].left_edge.i == 3
    assert doc[4].right_edge.i == 7

    assert doc[3].n_lefts == 0
    assert doc[3].n_rights == 0
    assert doc[3].left_edge.i == 3
    assert doc[3].right_edge.i == 3

    assert doc[2].left_edge.i == 0
    assert doc[2].right_edge.i == 8

    doc[6].head = doc[3]

    assert doc[6].n_lefts == 1
    assert doc[6].n_rights == 1
    assert doc[6].left_edge.i == 5
    assert doc[6].right_edge.i == 7

    assert doc[3].n_lefts == 0
    assert doc[3].n_rights == 1
    assert doc[3].left_edge.i == 3
    assert doc[3].right_edge.i == 7

    assert doc[4].n_lefts == 1
    assert doc[4].n_rights == 0
    assert doc[4].left_edge.i == 3
    assert doc[4].right_edge.i == 7

    assert doc[2].left_edge.i == 0
    assert doc[2].right_edge.i == 8

    doc[0].head = doc[5]

    assert doc[5].left_edge.i == 0
    assert doc[6].left_edge.i == 0
    assert doc[3].left_edge.i == 0
    assert doc[4].left_edge.i == 0
    assert doc[2].left_edge.i == 0

    # head token must be from the same document
    doc2 = get_doc(tokens.vocab, words=[t.text for t in tokens], heads=heads)
    with pytest.raises(ValueError):
        doc[0].head = doc2[0]


def test_is_sent_start(en_tokenizer):
    doc = en_tokenizer("This is a sentence. This is another.")
    assert doc[5].is_sent_start is None
    doc[5].is_sent_start = True
    assert doc[5].is_sent_start is True
    doc.is_parsed = True
    assert len(list(doc.sents)) == 2


def test_is_sent_end(en_tokenizer):
    doc = en_tokenizer("This is a sentence. This is another.")
    assert doc[4].is_sent_end is None
    doc[5].is_sent_start = True
    assert doc[4].is_sent_end is True
    doc.is_parsed = True
    assert len(list(doc.sents)) == 2


def test_set_pos():
    doc = Doc(Vocab(), words=["hello", "world"])
    doc[0].pos_ = "NOUN"
    assert doc[0].pos_ == "NOUN"
    doc[1].pos = VERB
    assert doc[1].pos_ == "VERB"


def test_tokens_sent(doc):
    """Test token.sent property"""
    assert len(list(doc.sents)) == 3
    assert doc[1].sent.text == "This is a sentence ."
    assert doc[7].sent.text == "This is another sentence ."
    assert doc[1].sent.root.left_edge.text == "This"
    assert doc[7].sent.root.left_edge.text == "This"


def test_token0_has_sent_start_true():
    doc = Doc(Vocab(), words=["hello", "world"])
    assert doc[0].is_sent_start is True
    assert doc[1].is_sent_start is None
    assert not doc.is_sentenced


def test_tokenlast_has_sent_end_true():
    doc = Doc(Vocab(), words=["hello", "world"])
    assert doc[0].is_sent_end is None
    assert doc[1].is_sent_end is True
    assert not doc.is_sentenced


def test_token_api_conjuncts_chain(en_vocab):
    words = "The boy and the girl and the man went .".split()
    heads = [1, 7, -1, 1, -3, -1, 1, -3, 0, -1]
    deps = ["det", "nsubj", "cc", "det", "conj", "cc", "det", "conj", "ROOT", "punct"]
    doc = get_doc(en_vocab, words=words, heads=heads, deps=deps)
    assert [w.text for w in doc[1].conjuncts] == ["girl", "man"]
    assert [w.text for w in doc[4].conjuncts] == ["boy", "man"]
    assert [w.text for w in doc[7].conjuncts] == ["boy", "girl"]


def test_token_api_conjuncts_simple(en_vocab):
    words = "They came and went .".split()
    heads = [1, 0, -1, -2, -1]
    deps = ["nsubj", "ROOT", "cc", "conj", "dep"]
    doc = get_doc(en_vocab, words=words, heads=heads, deps=deps)
    assert [w.text for w in doc[1].conjuncts] == ["went"]
    assert [w.text for w in doc[3].conjuncts] == ["came"]


def test_token_api_non_conjuncts(en_vocab):
    words = "They came .".split()
    heads = [1, 0, -1]
    deps = ["nsubj", "ROOT", "punct"]
    doc = get_doc(en_vocab, words=words, heads=heads, deps=deps)
    assert [w.text for w in doc[0].conjuncts] == []
    assert [w.text for w in doc[1].conjuncts] == []



from libcpp.vector cimport vector

from preshed.maps cimport PreshMap
from cymem.cymem cimport Pool
from murmurhash.mrmr cimport hash64

from .structs cimport LexemeC, TokenC
from .typedefs cimport attr_t, hash_t
from .strings cimport StringStore
from .morphology cimport Morphology


cdef LexemeC EMPTY_LEXEME


cdef union LexemesOrTokens:
    const LexemeC* const* lexemes
    const TokenC* tokens


cdef struct _Cached:
    LexemesOrTokens data
    bint is_lex
    int length


cdef class Vocab:
    cdef Pool mem
    cpdef readonly StringStore strings
    cpdef public Morphology morphology
    cpdef public object vectors
    cpdef public object lookups
    cpdef public object lookups_extra
    cdef readonly int length
    cdef public object data_dir
    cdef public object lex_attr_getters
    cdef public object cfg

    cdef const LexemeC* get(self, Pool mem, unicode string) except NULL
    cdef const LexemeC* get_by_orth(self, Pool mem, attr_t orth) except NULL
    cdef const TokenC* make_fused_token(self, substrings) except NULL

    cdef const LexemeC* _new_lexeme(self, Pool mem, unicode string) except NULL
    cdef int _add_lex_to_vocab(self, hash_t key, const LexemeC* lex) except -1
    cdef const LexemeC* _new_lexeme(self, Pool mem, unicode string) except NULL

    cdef PreshMap _by_orth



# coding: utf8
from __future__ import unicode_literals

from .stop_words import STOP_WORDS
from .tag_map import TAG_MAP
from .lex_attrs import LEX_ATTRS

from ...language import Language
from ...attrs import LANG


class SlovakDefaults(Language.Defaults):
    lex_attr_getters = dict(Language.Defaults.lex_attr_getters)
    lex_attr_getters.update(LEX_ATTRS)
    lex_attr_getters[LANG] = lambda text: "sk"
    tag_map = TAG_MAP
    stop_words = STOP_WORDS


class Slovak(Language):
    lang = "sk"
    Defaults = SlovakDefaults


__all__ = ["Slovak"]



# coding: utf8
from __future__ import unicode_literals

import pytest
import random
import numpy.random
from spacy.language import Language
from spacy.pipeline import TextCategorizer
from spacy.tokens import Doc
from spacy.gold import GoldParse


@pytest.mark.skip(reason="Test is flakey when run with others")
def test_simple_train():
    nlp = Language()
    nlp.add_pipe(nlp.create_pipe("textcat"))
    nlp.get_pipe("textcat").add_label("answer")
    nlp.begin_training()
    for i in range(5):
        for text, answer in [
            ("aaaa", 1.0),
            ("bbbb", 0),
            ("aa", 1.0),
            ("bbbbbbbbb", 0.0),
            ("aaaaaa", 1),
        ]:
            nlp.update([text], [{"cats": {"answer": answer}}])
    doc = nlp("aaa")
    assert "answer" in doc.cats
    assert doc.cats["answer"] >= 0.5


@pytest.mark.skip(reason="Test is flakey when run with others")
def test_textcat_learns_multilabel():
    random.seed(5)
    numpy.random.seed(5)
    docs = []
    nlp = Language()
    letters = ["a", "b", "c"]
    for w1 in letters:
        for w2 in letters:
            cats = {letter: float(w2 == letter) for letter in letters}
            docs.append((Doc(nlp.vocab, words=["d"] * 3 + [w1, w2] + ["d"] * 3), cats))
    random.shuffle(docs)
    model = TextCategorizer(nlp.vocab, width=8)
    for letter in letters:
        model.add_label(letter)
    optimizer = model.begin_training()
    for i in range(30):
        losses = {}
        Ys = [GoldParse(doc, cats=cats) for doc, cats in docs]
        Xs = [doc for doc, cats in docs]
        model.update(Xs, Ys, sgd=optimizer, losses=losses)
        random.shuffle(docs)
    for w1 in letters:
        for w2 in letters:
            doc = Doc(nlp.vocab, words=["d"] * 3 + [w1, w2] + ["d"] * 3)
            truth = {letter: w2 == letter for letter in letters}
            model(doc)
            for cat, score in doc.cats.items():
                if not truth[cat]:
                    assert score < 0.5
                else:
                    assert score > 0.5


def test_label_types():
    nlp = Language()
    nlp.add_pipe(nlp.create_pipe("textcat"))
    nlp.get_pipe("textcat").add_label("answer")
    with pytest.raises(ValueError):
        nlp.get_pipe("textcat").add_label(9)



# coding: utf8
from __future__ import unicode_literals

from ...attrs import LIKE_NUM

_num_words = [
    "zero",
    "one",
    "two",
    "three",
    "four",
    "five",
    "six",
    "seven",
    "eight",
    "nine",
    "ten",
    "eleven",
    "twelve",
    "thirteen",
    "fourteen",
    "fifteen",
    "sixteen",
    "seventeen",
    "eighteen",
    "nineteen",
    "twenty",
    "thirty",
    "forty",
    "fifty",
    "sixty",
    "seventy",
    "eighty",
    "ninety",
    "hundred",
    "thousand",
    "million",
    "billion",
    "trillion",
    "quadrillion",
    "gajillion",
    "bazillion",
]


_ordinal_words = [
    "first",
    "second",
    "third",
    "fourth",
    "fifth",
    "sixth",
    "seventh",
    "eighth",
    "ninth",
    "tenth",
    "eleventh",
    "twelfth",
    "thirteenth",
    "fourteenth",
    "fifteenth",
    "sixteenth",
    "seventeenth",
    "eighteenth",
    "nineteenth",
    "twentieth",
    "thirtieth",
    "fortieth",
    "fiftieth",
    "sixtieth",
    "seventieth",
    "eightieth",
    "ninetieth",
    "hundredth",
    "thousandth",
    "millionth",
    "billionth",
    "trillionth",
    "quadrillionth",
    "gajillionth",
    "bazillionth",
]

def like_num(text):
    if text.startswith(("+", "-", "±", "~")):
        text = text[1:]
    text = text.replace(",", "").replace(".", "")
    if text.isdigit():
        return True
    if text.count("/") == 1:
        num, denom = text.split("/")
        if num.isdigit() and denom.isdigit():
            return True

    text_lower = text.lower()
    if text_lower in _num_words:
        return True

    # CHeck ordinal number
    if text_lower in _ordinal_words:
        return True
    if text_lower.endswith("th"):
        if text_lower[:-2].isdigit():
            return True 

    return False


LEX_ATTRS = {LIKE_NUM: like_num}



# coding: utf8
from __future__ import unicode_literals

from ...attrs import LIKE_NUM

_num_words = {
    "antra",
    "antrai",
    "antrais",
    "antram",
    "antrame",
    "antras",
    "antri",
    "antriems",
    "antro",
    "antroje",
    "antromis",
    "antroms",
    "antros",
    "antrose",
    "antru",
    "antruose",
    "antrus",
    "antrą",
    "antrų",
    "aštunta",
    "aštuntai",
    "aštuntais",
    "aštuntam",
    "aštuntame",
    "aštuntas",
    "aštunti",
    "aštuntiems",
    "aštunto",
    "aštuntoje",
    "aštuntomis",
    "aštuntoms",
    "aštuntos",
    "aštuntose",
    "aštuntu",
    "aštuntuose",
    "aštuntus",
    "aštuntą",
    "aštuntų",
    "aštuoneri",
    "aštuoneriais",
    "aštuonerias",
    "aštuoneriems",
    "aštuoneriomis",
    "aštuonerioms",
    "aštuonerios",
    "aštuoneriose",
    "aštuoneriuose",
    "aštuonerius",
    "aštuonerių",
    "aštuonetai",
    "aštuonetais",
    "aštuonetams",
    "aštuonetas",
    "aštuonete",
    "aštuoneto",
    "aštuonetu",
    "aštuonetui",
    "aštuonetuose",
    "aštuonetus",
    "aštuonetą",
    "aštuonetų",
    "aštuoni",
    "aštuoniais",
    "aštuonias",
    "aštuoniasdešimt",
    "aštuoniasdešimta",
    "aštuoniasdešimtai",
    "aštuoniasdešimtais",
    "aštuoniasdešimtam",
    "aštuoniasdešimtame",
    "aštuoniasdešimtas",
    "aštuoniasdešimti",
    "aštuoniasdešimtiems",
    "aštuoniasdešimties",
    "aštuoniasdešimtimi",
    "aštuoniasdešimtimis",
    "aštuoniasdešimtims",
    "aštuoniasdešimtis",
    "aštuoniasdešimto",
    "aštuoniasdešimtoje",
    "aštuoniasdešimtomis",
    "aštuoniasdešimtoms",
    "aštuoniasdešimtos",
    "aštuoniasdešimtose",
    "aštuoniasdešimtu",
    "aštuoniasdešimtuose",
    "aštuoniasdešimtus",
    "aštuoniasdešimtyje",
    "aštuoniasdešimtys",
    "aštuoniasdešimtyse",
    "aštuoniasdešimtą",
    "aštuoniasdešimtį",
    "aštuoniasdešimtų",
    "aštuoniasdešimčia",
    "aštuoniasdešimčiai",
    "aštuoniasdešimčių",
    "aštuoniems",
    "aštuoniolika",
    "aštuoniolikai",
    "aštuoniolikoje",
    "aštuoniolikos",
    "aštuoniolikta",
    "aštuonioliktai",
    "aštuonioliktais",
    "aštuonioliktam",
    "aštuonioliktame",
    "aštuonioliktas",
    "aštuoniolikti",
    "aštuonioliktiems",
    "aštuoniolikto",
    "aštuonioliktoje",
    "aštuonioliktomis",
    "aštuonioliktoms",
    "aštuonioliktos",
    "aštuonioliktose",
    "aštuonioliktu",
    "aštuonioliktuose",
    "aštuonioliktus",
    "aštuonioliktą",
    "aštuonioliktų",
    "aštuoniomis",
    "aštuonioms",
    "aštuonios",
    "aštuoniose",
    "aštuonis",
    "aštuoniuose",
    "aštuonių",
    "bilijonai",
    "bilijonais",
    "bilijonams",
    "bilijonas",
    "bilijone",
    "bilijono",
    "bilijonu",
    "bilijonui",
    "bilijonuose",
    "bilijonus",
    "bilijoną",
    "bilijonų",
    "devinta",
    "devintai",
    "devintais",
    "devintam",
    "devintame",
    "devintas",
    "devinti",
    "devintiems",
    "devinto",
    "devintoje",
    "devintomis",
    "devintoms",
    "devintos",
    "devintose",
    "devintu",
    "devintuose",
    "devintus",
    "devintą",
    "devintų",
    "devyneri",
    "devyneriais",
    "devynerias",
    "devyneriems",
    "devyneriomis",
    "devynerioms",
    "devynerios",
    "devyneriose",
    "devyneriuose",
    "devynerius",
    "devynerių",
    "devynetas",
    "devynete",
    "devyneto",
    "devynetu",
    "devynetui",
    "devynetą",
    "devyni",
    "devyniais",
    "devynias",
    "devyniasdešimt",
    "devyniasdešimta",
    "devyniasdešimtai",
    "devyniasdešimtais",
    "devyniasdešimtam",
    "devyniasdešimtame",
    "devyniasdešimtas",
    "devyniasdešimti",
    "devyniasdešimtiems",
    "devyniasdešimties",
    "devyniasdešimtimi",
    "devyniasdešimtimis",
    "devyniasdešimtims",
    "devyniasdešimtis",
    "devyniasdešimto",
    "devyniasdešimtoje",
    "devyniasdešimtomis",
    "devyniasdešimtoms",
    "devyniasdešimtos",
    "devyniasdešimtose",
    "devyniasdešimtu",
    "devyniasdešimtuose",
    "devyniasdešimtus",
    "devyniasdešimtyje",
    "devyniasdešimtys",
    "devyniasdešimtyse",
    "devyniasdešimtą",
    "devyniasdešimtį",
    "devyniasdešimtų",
    "devyniasdešimčia",
    "devyniasdešimčiai",
    "devyniasdešimčių",
    "devyniems",
    "devyniolika",
    "devyniolikai",
    "devyniolikoje",
    "devyniolikos",
    "devyniolikta",
    "devynioliktai",
    "devynioliktais",
    "devynioliktam",
    "devynioliktame",
    "devynioliktas",
    "devyniolikti",
    "devynioliktiems",
    "devyniolikto",
    "devynioliktoje",
    "devynioliktomis",
    "devynioliktoms",
    "devynioliktos",
    "devynioliktose",
    "devynioliktu",
    "devynioliktuose",
    "devynioliktus",
    "devynioliktą",
    "devynioliktų",
    "devyniomis",
    "devynioms",
    "devynios",
    "devyniose",
    "devynis",
    "devyniuose",
    "devynių",
    "dešimt",
    "dešimta",
    "dešimtai",
    "dešimtais",
    "dešimtam",
    "dešimtame",
    "dešimtas",
    "dešimti",
    "dešimtiems",
    "dešimties",
    "dešimtimi",
    "dešimtimis",
    "dešimtims",
    "dešimtis",
    "dešimto",
    "dešimtoje",
    "dešimtomis",
    "dešimtoms",
    "dešimtos",
    "dešimtose",
    "dešimtu",
    "dešimtuose",
    "dešimtus",
    "dešimtyje",
    "dešimtys",
    "dešimtyse",
    "dešimtą",
    "dešimtį",
    "dešimtų",
    "dešimčia",
    "dešimčiai",
    "dešimčių",
    "du",
    "dutūkstanta",
    "dutūkstantai",
    "dutūkstantais",
    "dutūkstantam",
    "dutūkstantame",
    "dutūkstantas",
    "dutūkstanti",
    "dutūkstantiems",
    "dutūkstanto",
    "dutūkstantoje",
    "dutūkstantomis",
    "dutūkstantoms",
    "dutūkstantos",
    "dutūkstantose",
    "dutūkstantu",
    "dutūkstantuose",
    "dutūkstantus",
    "dutūkstantą",
    "dutūkstantų",
    "dušimta",
    "dušimtai",
    "dušimtais",
    "dušimtam",
    "dušimtame",
    "dušimtas",
    "dušimti",
    "dušimtiems",
    "dušimto",
    "dušimtoje",
    "dušimtomis",
    "dušimtoms",
    "dušimtos",
    "dušimtose",
    "dušimtu",
    "dušimtuose",
    "dušimtus",
    "dušimtą",
    "dušimtų",
    "dvejais",
    "dvejas",
    "dvejetai",
    "dvejetais",
    "dvejetams",
    "dvejetas",
    "dvejete",
    "dvejeto",
    "dvejetu",
    "dvejetui",
    "dvejetuose",
    "dvejetus",
    "dvejetą",
    "dvejetų",
    "dveji",
    "dvejiems",
    "dvejomis",
    "dvejoms",
    "dvejos",
    "dvejose",
    "dvejuose",
    "dvejus",
    "dvejų",
    "dvi",
    "dvidešimt",
    "dvidešimta",
    "dvidešimtai",
    "dvidešimtais",
    "dvidešimtam",
    "dvidešimtame",
    "dvidešimtas",
    "dvidešimti",
    "dvidešimtiems",
    "dvidešimties",
    "dvidešimtimi",
    "dvidešimtimis",
    "dvidešimtims",
    "dvidešimtis",
    "dvidešimto",
    "dvidešimtoje",
    "dvidešimtomis",
    "dvidešimtoms",
    "dvidešimtos",
    "dvidešimtose",
    "dvidešimtu",
    "dvidešimtuose",
    "dvidešimtus",
    "dvidešimtyje",
    "dvidešimtys",
    "dvidešimtyse",
    "dvidešimtą",
    "dvidešimtį",
    "dvidešimtų",
    "dvidešimčia",
    "dvidešimčiai",
    "dvidešimčių",
    "dviejose",
    "dviejuose",
    "dviejų",
    "dviem",
    "dvylika",
    "dvylikai",
    "dvylikoje",
    "dvylikos",
    "dvylikta",
    "dvyliktai",
    "dvyliktais",
    "dvyliktam",
    "dvyliktame",
    "dvyliktas",
    "dvylikti",
    "dvyliktiems",
    "dvylikto",
    "dvyliktoje",
    "dvyliktomis",
    "dvyliktoms",
    "dvyliktos",
    "dvyliktose",
    "dvyliktu",
    "dvyliktuose",
    "dvyliktus",
    "dvyliktą",
    "dvyliktų",
    "keturi",
    "keturiais",
    "keturias",
    "keturiasdešimt",
    "keturiasdešimta",
    "keturiasdešimtai",
    "keturiasdešimtais",
    "keturiasdešimtam",
    "keturiasdešimtame",
    "keturiasdešimtas",
    "keturiasdešimti",
    "keturiasdešimtiems",
    "keturiasdešimties",
    "keturiasdešimtimi",
    "keturiasdešimtimis",
    "keturiasdešimtims",
    "keturiasdešimtis",
    "keturiasdešimto",
    "keturiasdešimtoje",
    "keturiasdešimtomis",
    "keturiasdešimtoms",
    "keturiasdešimtos",
    "keturiasdešimtose",
    "keturiasdešimtu",
    "keturiasdešimtuose",
    "keturiasdešimtus",
    "keturiasdešimtyje",
    "keturiasdešimtys",
    "keturiasdešimtyse",
    "keturiasdešimtą",
    "keturiasdešimtį",
    "keturiasdešimtų",
    "keturiasdešimčia",
    "keturiasdešimčiai",
    "keturiasdešimčių",
    "keturiems",
    "keturiolika",
    "keturiolikai",
    "keturiolikoje",
    "keturiolikos",
    "keturiolikta",
    "keturioliktai",
    "keturioliktais",
    "keturioliktam",
    "keturioliktame",
    "keturioliktas",
    "keturiolikti",
    "keturioliktiems",
    "keturiolikto",
    "keturioliktoje",
    "keturioliktomis",
    "keturioliktoms",
    "keturioliktos",
    "keturioliktose",
    "keturioliktu",
    "keturioliktuose",
    "keturioliktus",
    "keturioliktą",
    "keturioliktų",
    "keturiomis",
    "keturioms",
    "keturios",
    "keturiose",
    "keturis",
    "keturiuose",
    "keturių",
    "ketveri",
    "ketveriais",
    "ketverias",
    "ketveriems",
    "ketveriomis",
    "ketverioms",
    "ketverios",
    "ketveriose",
    "ketveriuose",
    "ketverius",
    "ketverių",
    "ketvertai",
    "ketvertais",
    "ketvertams",
    "ketvertas",
    "ketverte",
    "ketverto",
    "ketvertu",
    "ketvertui",
    "ketvertuose",
    "ketvertus",
    "ketvertą",
    "ketvertų",
    "ketvirta",
    "ketvirtai",
    "ketvirtais",
    "ketvirtam",
    "ketvirtame",
    "ketvirtas",
    "ketvirti",
    "ketvirtiems",
    "ketvirto",
    "ketvirtoje",
    "ketvirtomis",
    "ketvirtoms",
    "ketvirtos",
    "ketvirtose",
    "ketvirtu",
    "ketvirtuose",
    "ketvirtus",
    "ketvirtą",
    "ketvirtų",
    "milijardai",
    "milijardais",
    "milijardams",
    "milijardas",
    "milijarde",
    "milijardo",
    "milijardu",
    "milijardui",
    "milijarduose",
    "milijardus",
    "milijardą",
    "milijardų",
    "milijonai",
    "milijonais",
    "milijonams",
    "milijonas",
    "milijone",
    "milijono",
    "milijonu",
    "milijonui",
    "milijonuose",
    "milijonus",
    "milijoną",
    "milijonų",
    "penkeri",
    "penkeriais",
    "penkerias",
    "penkeriems",
    "penkeriomis",
    "penkerioms",
    "penkerios",
    "penkeriose",
    "penkeriuose",
    "penkerius",
    "penkerių",
    "penketai",
    "penketais",
    "penketams",
    "penketas",
    "penkete",
    "penketo",
    "penketu",
    "penketui",
    "penketuose",
    "penketus",
    "penketą",
    "penketų",
    "penki",
    "penkiais",
    "penkias",
    "penkiasdešimt",
    "penkiasdešimta",
    "penkiasdešimtai",
    "penkiasdešimtais",
    "penkiasdešimtam",
    "penkiasdešimtame",
    "penkiasdešimtas",
    "penkiasdešimti",
    "penkiasdešimtiems",
    "penkiasdešimties",
    "penkiasdešimtimi",
    "penkiasdešimtimis",
    "penkiasdešimtims",
    "penkiasdešimtis",
    "penkiasdešimto",
    "penkiasdešimtoje",
    "penkiasdešimtomis",
    "penkiasdešimtoms",
    "penkiasdešimtos",
    "penkiasdešimtose",
    "penkiasdešimtu",
    "penkiasdešimtuose",
    "penkiasdešimtus",
    "penkiasdešimtyje",
    "penkiasdešimtys",
    "penkiasdešimtyse",
    "penkiasdešimtą",
    "penkiasdešimtį",
    "penkiasdešimtų",
    "penkiasdešimčia",
    "penkiasdešimčiai",
    "penkiasdešimčių",
    "penkiems",
    "penkiolika",
    "penkiolikai",
    "penkiolikoje",
    "penkiolikos",
    "penkiolikta",
    "penkioliktai",
    "penkioliktais",
    "penkioliktam",
    "penkioliktame",
    "penkioliktas",
    "penkiolikti",
    "penkioliktiems",
    "penkiolikto",
    "penkioliktoje",
    "penkioliktomis",
    "penkioliktoms",
    "penkioliktos",
    "penkioliktose",
    "penkioliktu",
    "penkioliktuose",
    "penkioliktus",
    "penkioliktą",
    "penkioliktų",
    "penkiomis",
    "penkioms",
    "penkios",
    "penkiose",
    "penkis",
    "penkiuose",
    "penkišimta",
    "penkišimtai",
    "penkišimtais",
    "penkišimtam",
    "penkišimtame",
    "penkišimtas",
    "penkišimti",
    "penkišimtiems",
    "penkišimto",
    "penkišimtoje",
    "penkišimtomis",
    "penkišimtoms",
    "penkišimtos",
    "penkišimtose",
    "penkišimtu",
    "penkišimtuose",
    "penkišimtus",
    "penkišimtą",
    "penkišimtų",
    "penkių",
    "penkta",
    "penktai",
    "penktais",
    "penktam",
    "penktame",
    "penktas",
    "penkti",
    "penktiems",
    "penkto",
    "penktoje",
    "penktomis",
    "penktoms",
    "penktos",
    "penktose",
    "penktu",
    "penktuose",
    "penktus",
    "penktą",
    "penktų",
    "pirma",
    "pirmai",
    "pirmais",
    "pirmaisiais",
    "pirmajai",
    "pirmajam",
    "pirmajame",
    "pirmam",
    "pirmame",
    "pirmas",
    "pirmasis",
    "pirmi",
    "pirmieji",
    "pirmiems",
    "pirmiesiems",
    "pirmo",
    "pirmoje",
    "pirmoji",
    "pirmojo",
    "pirmojoje",
    "pirmomis",
    "pirmoms",
    "pirmos",
    "pirmose",
    "pirmosiomis",
    "pirmosioms",
    "pirmosios",
    "pirmosiose",
    "pirmu",
    "pirmuoju",
    "pirmuose",
    "pirmuosiuose",
    "pirmuosius",
    "pirmus",
    "pirmą",
    "pirmąja",
    "pirmąją",
    "pirmąjį",
    "pirmąsias",
    "pirmų",
    "pirmųjų",
    "pusantro",
    "pusantros",
    "pusantrų",
    "pusdevinto",
    "pusdevintos",
    "pusdevintų",
    "pusdešimto",
    "pusdešimtos",
    "pusdešimtų",
    "pusketvirto",
    "pusketvirtos",
    "pusketvirtų",
    "puspenkto",
    "puspenktos",
    "puspenktų",
    "pusseptinto",
    "pusseptintos",
    "pusseptintų",
    "pustrečio",
    "pustrečios",
    "pustrečių",
    "pusšešto",
    "pusšeštos",
    "pusšeštų",
    "septinta",
    "septintai",
    "septintais",
    "septintam",
    "septintame",
    "septintas",
    "septinti",
    "septintiems",
    "septinto",
    "septintoje",
    "septintomis",
    "septintoms",
    "septintos",
    "septintose",
    "septintu",
    "septintuose",
    "septintus",
    "septintą",
    "septintų",
    "septyneri",
    "septyneriais",
    "septynerias",
    "septyneriems",
    "septyneriomis",
    "septynerioms",
    "septynerios",
    "septyneriose",
    "septyneriuose",
    "septynerius",
    "septynerių",
    "septynetas",
    "septynete",
    "septyneto",
    "septynetu",
    "septynetui",
    "septynetą",
    "septyni",
    "septyniais",
    "septynias",
    "septyniasdešimt",
    "septyniasdešimta",
    "septyniasdešimtai",
    "septyniasdešimtais",
    "septyniasdešimtam",
    "septyniasdešimtame",
    "septyniasdešimtas",
    "septyniasdešimti",
    "septyniasdešimtiems",
    "septyniasdešimties",
    "septyniasdešimtimi",
    "septyniasdešimtimis",
    "septyniasdešimtims",
    "septyniasdešimtis",
    "septyniasdešimto",
    "septyniasdešimtoje",
    "septyniasdešimtomis",
    "septyniasdešimtoms",
    "septyniasdešimtos",
    "septyniasdešimtose",
    "septyniasdešimtu",
    "septyniasdešimtuose",
    "septyniasdešimtus",
    "septyniasdešimtyje",
    "septyniasdešimtys",
    "septyniasdešimtyse",
    "septyniasdešimtą",
    "septyniasdešimtį",
    "septyniasdešimtų",
    "septyniasdešimčia",
    "septyniasdešimčiai",
    "septyniasdešimčių",
    "septyniems",
    "septyniolika",
    "septyniolikai",
    "septyniolikoje",
    "septyniolikos",
    "septyniolikta",
    "septynioliktai",
    "septynioliktais",
    "septynioliktam",
    "septynioliktame",
    "septynioliktas",
    "septyniolikti",
    "septynioliktiems",
    "septyniolikto",
    "septynioliktoje",
    "septynioliktomis",
    "septynioliktoms",
    "septynioliktos",
    "septynioliktose",
    "septynioliktu",
    "septynioliktuose",
    "septynioliktus",
    "septynioliktą",
    "septynioliktų",
    "septyniomis",
    "septynioms",
    "septynios",
    "septyniose",
    "septynis",
    "septyniuose",
    "septynių",
    "skaitvardi",
    "skaitvardis",
    "skaitvardyje",
    "skaitvardį",
    "skaitvardžiai",
    "skaitvardžiais",
    "skaitvardžiams",
    "skaitvardžio",
    "skaitvardžiu",
    "skaitvardžiui",
    "skaitvardžiuose",
    "skaitvardžius",
    "skaitvardžių",
    "trejais",
    "trejas",
    "trejetai",
    "trejetais",
    "trejetams",
    "trejetas",
    "trejete",
    "trejeto",
    "trejetu",
    "trejetui",
    "trejetuose",
    "trejetus",
    "trejetą",
    "trejetų",
    "treji",
    "trejiems",
    "trejomis",
    "trejoms",
    "trejos",
    "trejose",
    "trejuose",
    "trejus",
    "trejų",
    "treti",
    "tretiems",
    "trečia",
    "trečiai",
    "trečiais",
    "trečiam",
    "trečiame",
    "trečias",
    "trečio",
    "trečioje",
    "trečiomis",
    "trečioms",
    "trečios",
    "trečiose",
    "trečiu",
    "trečiuose",
    "trečius",
    "trečią",
    "trečių",
    "trijose",
    "trijuose",
    "trijų",
    "trilijonai",
    "trilijonais",
    "trilijonams",
    "trilijonas",
    "trilijone",
    "trilijono",
    "trilijonu",
    "trilijonui",
    "trilijonuose",
    "trilijonus",
    "trilijoną",
    "trilijonų",
    "trimis",
    "trims",
    "tris",
    "trisdešimt",
    "trisdešimta",
    "trisdešimtai",
    "trisdešimtais",
    "trisdešimtam",
    "trisdešimtame",
    "trisdešimtas",
    "trisdešimti",
    "trisdešimtiems",
    "trisdešimties",
    "trisdešimtimi",
    "trisdešimtimis",
    "trisdešimtims",
    "trisdešimtis",
    "trisdešimto",
    "trisdešimtoje",
    "trisdešimtomis",
    "trisdešimtoms",
    "trisdešimtos",
    "trisdešimtose",
    "trisdešimtu",
    "trisdešimtuose",
    "trisdešimtus",
    "trisdešimtyje",
    "trisdešimtys",
    "trisdešimtyse",
    "trisdešimtą",
    "trisdešimtį",
    "trisdešimtų",
    "trisdešimčia",
    "trisdešimčiai",
    "trisdešimčių",
    "trylika",
    "trylikai",
    "trylikoje",
    "trylikos",
    "trylikta",
    "tryliktai",
    "tryliktais",
    "tryliktam",
    "tryliktame",
    "tryliktas",
    "trylikti",
    "tryliktiems",
    "trylikto",
    "tryliktoje",
    "tryliktomis",
    "tryliktoms",
    "tryliktos",
    "tryliktose",
    "tryliktu",
    "tryliktuose",
    "tryliktus",
    "tryliktą",
    "tryliktų",
    "trys",
    "tūkstanta",
    "tūkstantai",
    "tūkstantais",
    "tūkstantam",
    "tūkstantame",
    "tūkstantas",
    "tūkstanti",
    "tūkstantiems",
    "tūkstantis",
    "tūkstanto",
    "tūkstantoje",
    "tūkstantomis",
    "tūkstantoms",
    "tūkstantos",
    "tūkstantose",
    "tūkstantu",
    "tūkstantuose",
    "tūkstantus",
    "tūkstantyje",
    "tūkstantą",
    "tūkstantį",
    "tūkstantų",
    "tūkstančiai",
    "tūkstančiais",
    "tūkstančiams",
    "tūkstančio",
    "tūkstančiu",
    "tūkstančiui",
    "tūkstančiuose",
    "tūkstančius",
    "tūkstančių",
    "vienam",
    "viename",
    "vienas",
    "vieno",
    "vienu",
    "vienuolika",
    "vienuolikai",
    "vienuolikoje",
    "vienuolikos",
    "vienuolikta",
    "vienuoliktais",
    "vienuoliktam",
    "vienuoliktame",
    "vienuoliktas",
    "vienuolikti",
    "vienuoliktiems",
    "vienuolikto",
    "vienuoliktoje",
    "vienuoliktomis",
    "vienuoliktoms",
    "vienuoliktos",
    "vienuoliktose",
    "vienuoliktu",
    "vienuoliktuose",
    "vienuoliktus",
    "vienuoliktą",
    "vienuoliktų",
    "vienuoliką",
    "vieną",
    "šešeri",
    "šešeriais",
    "šešerias",
    "šešeriems",
    "šešeriomis",
    "šešerioms",
    "šešerios",
    "šešeriose",
    "šešeriuose",
    "šešerius",
    "šešerių",
    "šešetai",
    "šešetais",
    "šešetams",
    "šešetas",
    "šešete",
    "šešeto",
    "šešetu",
    "šešetui",
    "šešetuose",
    "šešetus",
    "šešetą",
    "šešetų",
    "šeši",
    "šešiais",
    "šešias",
    "šešiasdešimt",
    "šešiasdešimta",
    "šešiasdešimtai",
    "šešiasdešimtais",
    "šešiasdešimtam",
    "šešiasdešimtame",
    "šešiasdešimtas",
    "šešiasdešimti",
    "šešiasdešimtiems",
    "šešiasdešimties",
    "šešiasdešimtimi",
    "šešiasdešimtimis",
    "šešiasdešimtims",
    "šešiasdešimtis",
    "šešiasdešimto",
    "šešiasdešimtoje",
    "šešiasdešimtomis",
    "šešiasdešimtoms",
    "šešiasdešimtos",
    "šešiasdešimtose",
    "šešiasdešimtu",
    "šešiasdešimtuose",
    "šešiasdešimtus",
    "šešiasdešimtyje",
    "šešiasdešimtys",
    "šešiasdešimtyse",
    "šešiasdešimtą",
    "šešiasdešimtį",
    "šešiasdešimtų",
    "šešiasdešimčia",
    "šešiasdešimčiai",
    "šešiasdešimčių",
    "šešiems",
    "šešiolika",
    "šešiolikai",
    "šešiolikoje",
    "šešiolikos",
    "šešiolikta",
    "šešioliktai",
    "šešioliktais",
    "šešioliktam",
    "šešioliktame",
    "šešioliktas",
    "šešiolikti",
    "šešioliktiems",
    "šešiolikto",
    "šešioliktoje",
    "šešioliktomis",
    "šešioliktoms",
    "šešioliktos",
    "šešioliktose",
    "šešioliktu",
    "šešioliktuose",
    "šešioliktus",
    "šešioliktą",
    "šešioliktų",
    "šešiomis",
    "šešioms",
    "šešios",
    "šešiose",
    "šešis",
    "šešiuose",
    "šešių",
    "šešta",
    "šeštai",
    "šeštais",
    "šeštam",
    "šeštame",
    "šeštas",
    "šešti",
    "šeštiems",
    "šešto",
    "šeštoje",
    "šeštomis",
    "šeštoms",
    "šeštos",
    "šeštose",
    "šeštu",
    "šeštuose",
    "šeštus",
    "šeštą",
    "šeštų",
    "šimtai",
    "šimtais",
    "šimtams",
    "šimtas",
    "šimte",
    "šimto",
    "šimtu",
    "šimtui",
    "šimtuose",
    "šimtus",
    "šimtą",
    "šimtų",
}


def like_num(text):
    text = text.replace(",", "").replace(".", "")
    if text.isdigit():
        return True
    if text.count("/") == 1:
        num, denom = text.split("/")
        if num.isdigit() and denom.isdigit():
            return True
    if text.lower() in _num_words:
        return True
    return False


LEX_ATTRS = {LIKE_NUM: like_num}



import os
import sys
import pytest
import spacy
import multiprocessing

if __name__ == "__main__":
    multiprocessing.freeze_support()

    PACKAGE_DIR = os.path.abspath(os.path.dirname((spacy.__file__)))
    sys.exit(pytest.main([PACKAGE_DIR]))


# coding: utf8
from __future__ import unicode_literals


"""
Example sentences to test spaCy and its language models.

>>> from spacy.lang.id.examples import sentences
>>> docs = nlp.pipe(sentences)
"""


sentences = [
    "Indonesia merupakan negara kepulauan yang kaya akan budaya.",
    "Berapa banyak warga yang dibutuhkan saat kerja bakti?",
    "Penyaluran pupuk berasal dari lima lokasi yakni Bontang, Kalimantan Timur, Surabaya, Banyuwangi, Semarang, dan Makassar.",
    "PT Pupuk Kaltim telah menyalurkan 274.707 ton pupuk bersubsidi ke wilayah penyaluran di 14 provinsi.",
    "Jakarta adalah kota besar yang nyaris tidak pernah tidur."
    "Kamu ada di mana semalam?",
    "Siapa yang membeli makanan ringan tersebut?",
    "Siapa presiden pertama Republik Indonesia?",
]



# coding: utf8
from __future__ import division, print_function, unicode_literals

import numpy as np

from .gold import tags_to_entities, GoldParse
from .errors import Errors


class PRFScore(object):
    """
    A precision / recall / F score
    """

    def __init__(self):
        self.tp = 0
        self.fp = 0
        self.fn = 0

    def score_set(self, cand, gold):
        self.tp += len(cand.intersection(gold))
        self.fp += len(cand - gold)
        self.fn += len(gold - cand)

    @property
    def precision(self):
        return self.tp / (self.tp + self.fp + 1e-100)

    @property
    def recall(self):
        return self.tp / (self.tp + self.fn + 1e-100)

    @property
    def fscore(self):
        p = self.precision
        r = self.recall
        return 2 * ((p * r) / (p + r + 1e-100))


class ROCAUCScore(object):
    """
    An AUC ROC score.
    """

    def __init__(self):
        self.golds = []
        self.cands = []
        self.saved_score = 0.0
        self.saved_score_at_len = 0

    def score_set(self, cand, gold):
        self.cands.append(cand)
        self.golds.append(gold)

    @property
    def score(self):
        if len(self.golds) == self.saved_score_at_len:
            return self.saved_score
        try:
            self.saved_score = _roc_auc_score(self.golds, self.cands)
        # catch ValueError: Only one class present in y_true.
        # ROC AUC score is not defined in that case.
        except ValueError:
            self.saved_score = -float("inf")
        self.saved_score_at_len = len(self.golds)
        return self.saved_score


class Scorer(object):
    """Compute evaluation scores."""

    def __init__(self, eval_punct=False, pipeline=None):
        """Initialize the Scorer.

        eval_punct (bool): Evaluate the dependency attachments to and from
            punctuation.
        RETURNS (Scorer): The newly created object.

        DOCS: https://spacy.io/api/scorer#init
        """
        self.tokens = PRFScore()
        self.sbd = PRFScore()
        self.unlabelled = PRFScore()
        self.labelled = PRFScore()
        self.labelled_per_dep = dict()
        self.tags = PRFScore()
        self.ner = PRFScore()
        self.ner_per_ents = dict()
        self.eval_punct = eval_punct
        self.textcat = None
        self.textcat_per_cat = dict()
        self.textcat_positive_label = None
        self.textcat_multilabel = False

        if pipeline:
            for name, model in pipeline:
                if name == "textcat":
                    self.textcat_positive_label = model.cfg.get("positive_label", None)
                    if self.textcat_positive_label:
                        self.textcat = PRFScore()
                    if not model.cfg.get("exclusive_classes", False):
                        self.textcat_multilabel = True
                        for label in model.cfg.get("labels", []):
                            self.textcat_per_cat[label] = ROCAUCScore()
                    else:
                        for label in model.cfg.get("labels", []):
                            self.textcat_per_cat[label] = PRFScore()

    @property
    def tags_acc(self):
        """RETURNS (float): Part-of-speech tag accuracy (fine grained tags,
            i.e. `Token.tag`).
        """
        return self.tags.fscore * 100

    @property
    def token_acc(self):
        """RETURNS (float): Tokenization accuracy."""
        return self.tokens.precision * 100

    @property
    def uas(self):
        """RETURNS (float): Unlabelled dependency score."""
        return self.unlabelled.fscore * 100

    @property
    def las(self):
        """RETURNS (float): Labelled dependency score."""
        return self.labelled.fscore * 100

    @property
    def las_per_type(self):
        """RETURNS (dict): Scores per dependency label.
        """
        return {
            k: {"p": v.precision * 100, "r": v.recall * 100, "f": v.fscore * 100}
            for k, v in self.labelled_per_dep.items()
        }

    @property
    def ents_p(self):
        """RETURNS (float): Named entity accuracy (precision)."""
        return self.ner.precision * 100

    @property
    def ents_r(self):
        """RETURNS (float): Named entity accuracy (recall)."""
        return self.ner.recall * 100

    @property
    def ents_f(self):
        """RETURNS (float): Named entity accuracy (F-score)."""
        return self.ner.fscore * 100

    @property
    def ents_per_type(self):
        """RETURNS (dict): Scores per entity label.
        """
        return {
            k: {"p": v.precision * 100, "r": v.recall * 100, "f": v.fscore * 100}
            for k, v in self.ner_per_ents.items()
        }

    @property
    def textcat_score(self):
        """RETURNS (float): f-score on positive label for binary exclusive,
        macro-averaged f-score for 3+ exclusive,
        macro-averaged AUC ROC score for multilabel (-1 if undefined)
        """
        if not self.textcat_multilabel:
            # binary multiclass
            if self.textcat_positive_label:
                return self.textcat.fscore * 100
            # other multiclass
            return (
                sum([score.fscore for label, score in self.textcat_per_cat.items()])
                / (len(self.textcat_per_cat) + 1e-100)
                * 100
            )
        # multilabel
        return max(
            sum([score.score for label, score in self.textcat_per_cat.items()])
            / (len(self.textcat_per_cat) + 1e-100),
            -1,
        )

    @property
    def textcats_per_cat(self):
        """RETURNS (dict): Scores per textcat label.
        """
        if not self.textcat_multilabel:
            return {
                k: {"p": v.precision * 100, "r": v.recall * 100, "f": v.fscore * 100}
                for k, v in self.textcat_per_cat.items()
            }
        return {
            k: {"roc_auc_score": max(v.score, -1)}
            for k, v in self.textcat_per_cat.items()
        }

    @property
    def scores(self):
        """RETURNS (dict): All scores with keys `uas`, `las`, `ents_p`,
            `ents_r`, `ents_f`, `tags_acc`, `token_acc`, and `textcat_score`.
        """
        return {
            "uas": self.uas,
            "las": self.las,
            "las_per_type": self.las_per_type,
            "ents_p": self.ents_p,
            "ents_r": self.ents_r,
            "ents_f": self.ents_f,
            "ents_per_type": self.ents_per_type,
            "tags_acc": self.tags_acc,
            "token_acc": self.token_acc,
            "textcat_score": self.textcat_score,
            "textcats_per_cat": self.textcats_per_cat,
        }

    def score(self, doc, gold, verbose=False, punct_labels=("p", "punct")):
        """Update the evaluation scores from a single Doc / GoldParse pair.

        doc (Doc): The predicted annotations.
        gold (GoldParse): The correct annotations.
        verbose (bool): Print debugging information.
        punct_labels (tuple): Dependency labels for punctuation. Used to
            evaluate dependency attachments to punctuation if `eval_punct` is
            `True`.

        DOCS: https://spacy.io/api/scorer#score
        """
        if len(doc) != len(gold):
            gold = GoldParse.from_annot_tuples(
                doc, zip(*gold.orig_annot), cats=gold.cats,
            )
        gold_deps = set()
        gold_deps_per_dep = {}
        gold_tags = set()
        gold_ents = set(tags_to_entities([annot[-1] for annot in gold.orig_annot]))
        for id_, word, tag, head, dep, ner in gold.orig_annot:
            gold_tags.add((id_, tag))
            if dep not in (None, "") and dep.lower() not in punct_labels:
                gold_deps.add((id_, head, dep.lower()))
                if dep.lower() not in self.labelled_per_dep:
                    self.labelled_per_dep[dep.lower()] = PRFScore()
                if dep.lower() not in gold_deps_per_dep:
                    gold_deps_per_dep[dep.lower()] = set()
                gold_deps_per_dep[dep.lower()].add((id_, head, dep.lower()))
        cand_deps = set()
        cand_deps_per_dep = {}
        cand_tags = set()
        for token in doc:
            if token.orth_.isspace():
                continue
            gold_i = gold.cand_to_gold[token.i]
            if gold_i is None:
                self.tokens.fp += 1
            else:
                self.tokens.tp += 1
                cand_tags.add((gold_i, token.tag_))
            if token.dep_.lower() not in punct_labels and token.orth_.strip():
                gold_head = gold.cand_to_gold[token.head.i]
                # None is indistinct, so we can't just add it to the set
                # Multiple (None, None) deps are possible
                if gold_i is None or gold_head is None:
                    self.unlabelled.fp += 1
                    self.labelled.fp += 1
                else:
                    cand_deps.add((gold_i, gold_head, token.dep_.lower()))
                    if token.dep_.lower() not in self.labelled_per_dep:
                        self.labelled_per_dep[token.dep_.lower()] = PRFScore()
                    if token.dep_.lower() not in cand_deps_per_dep:
                        cand_deps_per_dep[token.dep_.lower()] = set()
                    cand_deps_per_dep[token.dep_.lower()].add(
                        (gold_i, gold_head, token.dep_.lower())
                    )
        if "-" not in [token[-1] for token in gold.orig_annot]:
            # Find all NER labels in gold and doc
            ent_labels = set([x[0] for x in gold_ents] + [k.label_ for k in doc.ents])
            # Set up all labels for per type scoring and prepare gold per type
            gold_per_ents = {ent_label: set() for ent_label in ent_labels}
            for ent_label in ent_labels:
                if ent_label not in self.ner_per_ents:
                    self.ner_per_ents[ent_label] = PRFScore()
                gold_per_ents[ent_label].update(
                    [x for x in gold_ents if x[0] == ent_label]
                )
            # Find all candidate labels, for all and per type
            cand_ents = set()
            cand_per_ents = {ent_label: set() for ent_label in ent_labels}
            for ent in doc.ents:
                first = gold.cand_to_gold[ent.start]
                last = gold.cand_to_gold[ent.end - 1]
                if first is None or last is None:
                    self.ner.fp += 1
                    self.ner_per_ents[ent.label_].fp += 1
                else:
                    cand_ents.add((ent.label_, first, last))
                    cand_per_ents[ent.label_].add((ent.label_, first, last))
            # Scores per ent
            for k, v in self.ner_per_ents.items():
                if k in cand_per_ents:
                    v.score_set(cand_per_ents[k], gold_per_ents[k])
            # Score for all ents
            self.ner.score_set(cand_ents, gold_ents)
        self.tags.score_set(cand_tags, gold_tags)
        self.labelled.score_set(cand_deps, gold_deps)
        for dep in self.labelled_per_dep:
            self.labelled_per_dep[dep].score_set(
                cand_deps_per_dep.get(dep, set()), gold_deps_per_dep.get(dep, set())
            )
        self.unlabelled.score_set(
            set(item[:2] for item in cand_deps), set(item[:2] for item in gold_deps)
        )
        if (
            len(gold.cats) > 0
            and set(self.textcat_per_cat) == set(gold.cats)
            and set(gold.cats) == set(doc.cats)
        ):
            goldcat = max(gold.cats, key=gold.cats.get)
            candcat = max(doc.cats, key=doc.cats.get)
            if self.textcat_positive_label:
                self.textcat.score_set(
                    set([self.textcat_positive_label]) & set([candcat]),
                    set([self.textcat_positive_label]) & set([goldcat]),
                )
            for label in self.textcat_per_cat:
                if self.textcat_multilabel:
                    self.textcat_per_cat[label].score_set(
                        doc.cats[label], gold.cats[label]
                    )
                else:
                    self.textcat_per_cat[label].score_set(
                        set([label]) & set([candcat]), set([label]) & set([goldcat])
                    )
        elif len(self.textcat_per_cat) > 0:
            model_labels = set(self.textcat_per_cat)
            eval_labels = set(gold.cats)
            raise ValueError(
                Errors.E162.format(model_labels=model_labels, eval_labels=eval_labels)
            )
        if verbose:
            gold_words = [item[1] for item in gold.orig_annot]
            for w_id, h_id, dep in cand_deps - gold_deps:
                print("F", gold_words[w_id], dep, gold_words[h_id])
            for w_id, h_id, dep in gold_deps - cand_deps:
                print("M", gold_words[w_id], dep, gold_words[h_id])


#############################################################################
#
# The following implementation of roc_auc_score() is adapted from
# scikit-learn, which is distributed under the following license:
#
# New BSD License
#
# Copyright (c) 2007–2019 The scikit-learn developers.
# All rights reserved.
#
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the Scikit-learn Developers  nor the names of
#      its contributors may be used to endorse or promote products
#      derived from this software without specific prior written
#      permission.
#
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.


def _roc_auc_score(y_true, y_score):
    """Compute Area Under the Receiver Operating Characteristic Curve (ROC AUC)
    from prediction scores.

    Note: this implementation is restricted to the binary classification task

    Parameters
    ----------
    y_true : array, shape = [n_samples] or [n_samples, n_classes]
        True binary labels or binary label indicators.
        The multiclass case expects shape = [n_samples] and labels
        with values in ``range(n_classes)``.

    y_score : array, shape = [n_samples] or [n_samples, n_classes]
        Target scores, can either be probability estimates of the positive
        class, confidence values, or non-thresholded measure of decisions
        (as returned by "decision_function" on some classifiers). For binary
        y_true, y_score is supposed to be the score of the class with greater
        label. The multiclass case expects shape = [n_samples, n_classes]
        where the scores correspond to probability estimates.

    Returns
    -------
    auc : float

    References
    ----------
    .. [1] `Wikipedia entry for the Receiver operating characteristic
            <https://en.wikipedia.org/wiki/Receiver_operating_characteristic>`_

    .. [2] Fawcett T. An introduction to ROC analysis[J]. Pattern Recognition
           Letters, 2006, 27(8):861-874.

    .. [3] `Analyzing a portion of the ROC curve. McClish, 1989
            <https://www.ncbi.nlm.nih.gov/pubmed/2668680>`_
    """
    if len(np.unique(y_true)) != 2:
        raise ValueError(Errors.E165)
    fpr, tpr, _ = _roc_curve(y_true, y_score)
    return _auc(fpr, tpr)


def _roc_curve(y_true, y_score):
    """Compute Receiver operating characteristic (ROC)

    Note: this implementation is restricted to the binary classification task.

    Parameters
    ----------

    y_true : array, shape = [n_samples]
        True binary labels. If labels are not either {-1, 1} or {0, 1}, then
        pos_label should be explicitly given.

    y_score : array, shape = [n_samples]
        Target scores, can either be probability estimates of the positive
        class, confidence values, or non-thresholded measure of decisions
        (as returned by "decision_function" on some classifiers).

    Returns
    -------
    fpr : array, shape = [>2]
        Increasing false positive rates such that element i is the false
        positive rate of predictions with score >= thresholds[i].

    tpr : array, shape = [>2]
        Increasing true positive rates such that element i is the true
        positive rate of predictions with score >= thresholds[i].

    thresholds : array, shape = [n_thresholds]
        Decreasing thresholds on the decision function used to compute
        fpr and tpr. `thresholds[0]` represents no instances being predicted
        and is arbitrarily set to `max(y_score) + 1`.

    Notes
    -----
    Since the thresholds are sorted from low to high values, they
    are reversed upon returning them to ensure they correspond to both ``fpr``
    and ``tpr``, which are sorted in reversed order during their calculation.

    References
    ----------
    .. [1] `Wikipedia entry for the Receiver operating characteristic
            <https://en.wikipedia.org/wiki/Receiver_operating_characteristic>`_

    .. [2] Fawcett T. An introduction to ROC analysis[J]. Pattern Recognition
           Letters, 2006, 27(8):861-874.
    """
    fps, tps, thresholds = _binary_clf_curve(y_true, y_score)

    # Add an extra threshold position
    # to make sure that the curve starts at (0, 0)
    tps = np.r_[0, tps]
    fps = np.r_[0, fps]
    thresholds = np.r_[thresholds[0] + 1, thresholds]

    if fps[-1] <= 0:
        fpr = np.repeat(np.nan, fps.shape)
    else:
        fpr = fps / fps[-1]

    if tps[-1] <= 0:
        tpr = np.repeat(np.nan, tps.shape)
    else:
        tpr = tps / tps[-1]

    return fpr, tpr, thresholds


def _binary_clf_curve(y_true, y_score):
    """Calculate true and false positives per binary classification threshold.

    Parameters
    ----------
    y_true : array, shape = [n_samples]
        True targets of binary classification

    y_score : array, shape = [n_samples]
        Estimated probabilities or decision function

    Returns
    -------
    fps : array, shape = [n_thresholds]
        A count of false positives, at index i being the number of negative
        samples assigned a score >= thresholds[i]. The total number of
        negative samples is equal to fps[-1] (thus true negatives are given by
        fps[-1] - fps).

    tps : array, shape = [n_thresholds <= len(np.unique(y_score))]
        An increasing count of true positives, at index i being the number
        of positive samples assigned a score >= thresholds[i]. The total
        number of positive samples is equal to tps[-1] (thus false negatives
        are given by tps[-1] - tps).

    thresholds : array, shape = [n_thresholds]
        Decreasing score values.
    """
    pos_label = 1.0

    y_true = np.ravel(y_true)
    y_score = np.ravel(y_score)

    # make y_true a boolean vector
    y_true = y_true == pos_label

    # sort scores and corresponding truth values
    desc_score_indices = np.argsort(y_score, kind="mergesort")[::-1]
    y_score = y_score[desc_score_indices]
    y_true = y_true[desc_score_indices]
    weight = 1.0

    # y_score typically has many tied values. Here we extract
    # the indices associated with the distinct values. We also
    # concatenate a value for the end of the curve.
    distinct_value_indices = np.where(np.diff(y_score))[0]
    threshold_idxs = np.r_[distinct_value_indices, y_true.size - 1]

    # accumulate the true positives with decreasing threshold
    tps = _stable_cumsum(y_true * weight)[threshold_idxs]
    fps = 1 + threshold_idxs - tps
    return fps, tps, y_score[threshold_idxs]


def _stable_cumsum(arr, axis=None, rtol=1e-05, atol=1e-08):
    """Use high precision for cumsum and check that final value matches sum

    Parameters
    ----------
    arr : array-like
        To be cumulatively summed as flat
    axis : int, optional
        Axis along which the cumulative sum is computed.
        The default (None) is to compute the cumsum over the flattened array.
    rtol : float
        Relative tolerance, see ``np.allclose``
    atol : float
        Absolute tolerance, see ``np.allclose``
    """
    out = np.cumsum(arr, axis=axis, dtype=np.float64)
    expected = np.sum(arr, axis=axis, dtype=np.float64)
    if not np.all(
        np.isclose(
            out.take(-1, axis=axis), expected, rtol=rtol, atol=atol, equal_nan=True
        )
    ):
        raise ValueError(Errors.E163)
    return out


def _auc(x, y):
    """Compute Area Under the Curve (AUC) using the trapezoidal rule

    This is a general function, given points on a curve.  For computing the
    area under the ROC-curve, see :func:`roc_auc_score`.

    Parameters
    ----------
    x : array, shape = [n]
        x coordinates. These must be either monotonic increasing or monotonic
        decreasing.
    y : array, shape = [n]
        y coordinates.

    Returns
    -------
    auc : float
    """
    x = np.ravel(x)
    y = np.ravel(y)

    direction = 1
    dx = np.diff(x)
    if np.any(dx < 0):
        if np.all(dx <= 0):
            direction = -1
        else:
            raise ValueError(Errors.E164.format(x))

    area = direction * np.trapz(y, x)
    if isinstance(area, np.memmap):
        # Reductions such as .sum used internally in np.trapz do not return a
        # scalar by default for numpy.memmap instances contrary to
        # regular numpy.ndarray instances.
        area = area.dtype.type(area)
    return area



import spacy
from spacy.language import Language
from spacy.lang.en import English
from spacy.tests.util import make_tempdir


def test_issue5137():
    class MyComponent(object):
        name = "my_component"

        def __init__(self, nlp, **cfg):
            self.nlp = nlp
            self.categories = cfg.get("categories", "all_categories")

        def __call__(self, doc):
            pass

        def to_disk(self, path, **kwargs):
            pass

        def from_disk(self, path, **cfg):
            pass

    Language.factories["my_component"] = lambda nlp, **cfg: MyComponent(nlp, **cfg)

    nlp = English()
    nlp.add_pipe(nlp.create_pipe("my_component"))
    assert nlp.get_pipe("my_component").categories == "all_categories"

    with make_tempdir() as tmpdir:
        nlp.to_disk(tmpdir)
        nlp2 = spacy.load(tmpdir, categories="my_categories")
        assert nlp2.get_pipe("my_component").categories == "my_categories"



# coding: utf8
from __future__ import unicode_literals


STOP_WORDS = set(
    """
а
авај
ако
ал
али
арх
ау
ах
аха
ај
бар
би
била
били
било
бисмо
бисте
бих
бијасмо
бијасте
бијах
бијаху
бијаше
биће
близу
број
брр
буде
будимо
будите
буду
будући
бум
бућ
вам
вама
вас
ваша
ваше
вашим
вашима
ваљда
веома
вероватно
већ
већина
ви
видео
више
врло
врх
га
где
гиц
год
горе
гђекоје
да
дакле
дана
данас
дај
два
де
дедер
делимице
делимично
дем
до
добар
добити
довечер
докле
доле
донекле
досад
доскоро
дотад
дотле
дошао
доћи
другамо
другде
други
е
ево
ено
ето
ех
ехе
еј
желела
желеле
желели
желело
желех
желећи
жели
за
заиста
зар
затим
зато
захвалити
зашто
збиља
зимус
знати
зум
и
иде
из
изван
изволи
између
изнад
икада
икакав
икаква
икакве
икакви
икаквим
икаквима
икаквих
икакво
икаквог
икаквога
икаквом
икаквоме
икаквој
или
им
има
имам
имао
испод
их
ију
ићи
кад
када
кога
којекакав
којима
коју
кришом
лани
ли
мали
мањи
ме
мене
мени
ми
мимо
мисли
много
могу
мора
морао
мој
моја
моје
моји
моју
моћи
му
на
над
након
нам
нама
нас
наша
наше
нашег
наши
наћи
не
негде
нека
некад
неке
неког
неку
нема
немам
неко
неће
нећемо
нећете
нећеш
нећу
ни
никада
никога
никоје
никоји
никоју
нисам
ниси
нисте
нису
ништа
ниједан
но
о
ова
овако
овамо
овај
овде
ове
овим
овима
ово
овој
од
одмах
око
около
он
онај
оне
оним
онима
оном
оној
ону
осим
остали
отишао
па
пак
питати
по
поводом
под
подаље
пожељан
пожељна
поиздаље
поименце
понекад
попреко
поред
после
потаман
потрбушке
поуздано
почетак
поједини
правити
први
преко
према
прије
пут
пљус
радије
с
са
сав
сада
сам
само
сасвим
сва
сваки
сви
свим
свог
свом
свој
своја
своје
своју
сву
свугде
се
себе
себи
си
смети
смо
ствар
стварно
сте
су
сутра
та
таèно
тако
такође
тамо
твој
твоја
твоје
твоји
твоју
те
тебе
теби
ти
тима
то
томе
тој
ту
у
увек
увијек
уз
уза
узалуд
уздуж
узети
умало
унутра
употребити
упркос
учинио
учинити
хало
хвала
хеј
хм
хоп
хоће
хоћемо
хоћете
хоћеш
хоћу
хтедосте
хтедох
хтедоше
хтела
хтеле
хтели
хтео
хтејасмо
хтејасте
хтејаху
хура
често
чијем
чији
чијим
чијима
шиц
штагод
што
штогод
ја
је
један
једини
једна
једне
једни
једно
једном
јер
јесам
јеси
јесмо
јесу
јим
јој
ју
јуче
његова
његово
њезин
њезина
њезино
њему
њен
њим
њима
њихова
њихово
њој
њу
ће
ћемо
ћете
ћеш
ћу
""".split()
)



# coding: utf8
from __future__ import unicode_literals

from ...attrs import LIKE_NUM


_num_words = set(
    """
zero unu doi două trei patru cinci șase șapte opt nouă zece
unsprezece doisprezece douăsprezece treisprezece patrusprezece cincisprezece șaisprezece șaptesprezece optsprezece nouăsprezece
douăzeci treizeci patruzeci cincizeci șaizeci șaptezeci optzeci nouăzeci
sută mie milion miliard bilion trilion cvadrilion catralion cvintilion sextilion septilion enșpemii
""".split()
)

_ordinal_words = set(
    """
primul doilea treilea patrulea cincilea șaselea șaptelea optulea nouălea zecelea
prima doua treia patra cincia șasea șaptea opta noua zecea
unsprezecelea doisprezecelea treisprezecelea patrusprezecelea cincisprezecelea șaisprezecelea șaptesprezecelea optsprezecelea nouăsprezecelea
unsprezecea douăsprezecea treisprezecea patrusprezecea cincisprezecea șaisprezecea șaptesprezecea optsprezecea nouăsprezecea
douăzecilea treizecilea patruzecilea cincizecilea șaizecilea șaptezecilea optzecilea nouăzecilea sutălea
douăzecea treizecea patruzecea cincizecea șaizecea șaptezecea optzecea nouăzecea suta
miilea mielea mia milionulea milioana miliardulea miliardelea miliarda enșpemia
""".split()
)


def like_num(text):
    if text.startswith(("+", "-", "±", "~")):
        text = text[1:]
    text = text.replace(",", "").replace(".", "")
    if text.isdigit():
        return True
    if text.count("/") == 1:
        num, denom = text.split("/")
        if num.isdigit() and denom.isdigit():
            return True
    if text.lower() in _num_words:
        return True
    if text.lower() in _ordinal_words:
        return True
    return False


LEX_ATTRS = {LIKE_NUM: like_num}



# coding: utf8
from __future__ import unicode_literals

import pytest


@pytest.mark.parametrize("text", ["αριθ.", "τρισ.", "δισ.", "σελ."])
def test_el_tokenizer_handles_abbr(el_tokenizer, text):
    tokens = el_tokenizer(text)
    assert len(tokens) == 1


def test_el_tokenizer_handles_exc_in_text(el_tokenizer):
    text = "Στα 14 τρισ. δολάρια το κόστος από την άνοδο της στάθμης της θάλασσας."
    tokens = el_tokenizer(text)
    assert len(tokens) == 14
    assert tokens[2].text == "τρισ."



# coding: utf8
from __future__ import unicode_literals, print_function

from pathlib import Path
import sys
import requests
import srsly
from wasabi import msg

from ..compat import path2str
from ..util import get_data_path
from .. import about


def validate():
    """
    Validate that the currently installed version of spaCy is compatible
    with the installed models. Should be run after `pip install -U spacy`.
    """
    with msg.loading("Loading compatibility table..."):
        r = requests.get(about.__compatibility__)
        if r.status_code != 200:
            msg.fail(
                "Server error ({})".format(r.status_code),
                "Couldn't fetch compatibility table.",
                exits=1,
            )
    msg.good("Loaded compatibility table")
    compat = r.json()["spacy"]
    version = about.__version__
    version = version.rsplit(".dev", 1)[0]
    current_compat = compat.get(version)
    if not current_compat:
        msg.fail(
            "Can't find spaCy v{} in compatibility table".format(version),
            about.__compatibility__,
            exits=1,
        )
    all_models = set()
    for spacy_v, models in dict(compat).items():
        all_models.update(models.keys())
        for model, model_vs in models.items():
            compat[spacy_v][model] = [reformat_version(v) for v in model_vs]
    model_links = get_model_links(current_compat)
    model_pkgs = get_model_pkgs(current_compat, all_models)
    incompat_links = {l for l, d in model_links.items() if not d["compat"]}
    incompat_models = {d["name"] for _, d in model_pkgs.items() if not d["compat"]}
    incompat_models.update(
        [d["name"] for _, d in model_links.items() if not d["compat"]]
    )
    na_models = [m for m in incompat_models if m not in current_compat]
    update_models = [m for m in incompat_models if m in current_compat]
    spacy_dir = Path(__file__).parent.parent

    msg.divider("Installed models (spaCy v{})".format(about.__version__))
    msg.info("spaCy installation: {}".format(path2str(spacy_dir)))

    if model_links or model_pkgs:
        header = ("TYPE", "NAME", "MODEL", "VERSION", "")
        rows = []
        for name, data in model_pkgs.items():
            rows.append(get_model_row(current_compat, name, data, msg))
        for name, data in model_links.items():
            rows.append(get_model_row(current_compat, name, data, msg, "link"))
        msg.table(rows, header=header)
    else:
        msg.text("No models found in your current environment.", exits=0)
    if update_models:
        msg.divider("Install updates")
        msg.text("Use the following commands to update the model packages:")
        cmd = "python -m spacy download {}"
        print("\n".join([cmd.format(pkg) for pkg in update_models]) + "\n")
    if na_models:
        msg.text(
            "The following models are not available for spaCy "
            "v{}: {}".format(about.__version__, ", ".join(na_models))
        )
    if incompat_links:
        msg.text(
            "You may also want to overwrite the incompatible links using the "
            "`python -m spacy link` command with `--force`, or remove them "
            "from the data directory. "
            "Data path: {path}".format(path=path2str(get_data_path()))
        )
    if incompat_models or incompat_links:
        sys.exit(1)


def get_model_links(compat):
    links = {}
    data_path = get_data_path()
    if data_path:
        models = [p for p in data_path.iterdir() if is_model_path(p)]
        for model in models:
            meta_path = Path(model) / "meta.json"
            if not meta_path.exists():
                continue
            meta = srsly.read_json(meta_path)
            link = model.parts[-1]
            name = meta["lang"] + "_" + meta["name"]
            links[link] = {
                "name": name,
                "version": meta["version"],
                "compat": is_compat(compat, name, meta["version"]),
            }
    return links


def get_model_pkgs(compat, all_models):
    import pkg_resources

    pkgs = {}
    for pkg_name, pkg_data in pkg_resources.working_set.by_key.items():
        package = pkg_name.replace("-", "_")
        if package in all_models:
            version = pkg_data.version
            pkgs[pkg_name] = {
                "name": package,
                "version": version,
                "compat": is_compat(compat, package, version),
            }
    return pkgs


def get_model_row(compat, name, data, msg, model_type="package"):
    if data["compat"]:
        comp = msg.text("", color="green", icon="good", no_print=True)
        version = msg.text(data["version"], color="green", no_print=True)
    else:
        version = msg.text(data["version"], color="red", no_print=True)
        comp = "--> {}".format(compat.get(data["name"], ["n/a"])[0])
    return (model_type, name, data["name"], version, comp)


def is_model_path(model_path):
    exclude = ["cache", "pycache", "__pycache__"]
    name = model_path.parts[-1]
    return model_path.is_dir() and name not in exclude and not name.startswith(".")


def is_compat(compat, name, version):
    return name in compat and version in compat[name]


def reformat_version(version):
    """Hack to reformat old versions ending on '-alpha' to match pip format."""
    if version.endswith("-alpha"):
        return version.replace("-alpha", "a0")
    return version.replace("-alpha", "a")



# cython: infer_types=True
# coding: utf-8
from __future__ import unicode_literals

from cpython.ref cimport Py_INCREF
from cymem.cymem cimport Pool
from thinc.typedefs cimport weight_t
from thinc.extra.search cimport Beam
from collections import OrderedDict, Counter
import srsly

from . cimport _beam_utils
from ..tokens.doc cimport Doc
from ..structs cimport TokenC
from .stateclass cimport StateClass
from ..typedefs cimport attr_t
from ..errors import Errors
from .. import util


cdef weight_t MIN_SCORE = -90000


class OracleError(Exception):
    pass


cdef void* _init_state(Pool mem, int length, void* tokens) except NULL:
    cdef StateC* st = new StateC(<const TokenC*>tokens, length)
    return <void*>st


cdef int _del_state(Pool mem, void* state, void* x) except -1:
    cdef StateC* st = <StateC*>state
    del st


cdef class TransitionSystem:
    def __init__(self, StringStore string_table, labels_by_action=None, min_freq=None):
        self.mem = Pool()
        self.strings = string_table
        self.n_moves = 0
        self._size = 100

        self.c = <Transition*>self.mem.alloc(self._size, sizeof(Transition))

        self.labels = {}
        if labels_by_action:
            self.initialize_actions(labels_by_action, min_freq=min_freq)
        self.root_label = self.strings.add('ROOT')
        self.init_beam_state = _init_state
        self.del_beam_state = _del_state

    def __reduce__(self):
        return (self.__class__, (self.strings, self.labels), None, None)

    def init_batch(self, docs):
        cdef StateClass state
        states = []
        offset = 0
        for doc in docs:
            state = StateClass(doc, offset=offset)
            self.initialize_state(state.c)
            states.append(state)
            offset += len(doc)
        return states

    def init_beams(self, docs, beam_width, beam_density=0.):
        cdef Doc doc
        beams = []
        cdef int offset = 0

        # Doc objects might contain labels that we need to register actions for. We need to check for that
        # *before* we create any Beam objects, because the Beam object needs the correct number of
        # actions. It's sort of dumb, but the best way is to just call init_batch() -- that triggers the additions,
        # and it doesn't matter that we create and discard the state objects.
        self.init_batch(docs)

        for doc in docs:
            beam = Beam(self.n_moves, beam_width, min_density=beam_density)
            beam.initialize(self.init_beam_state, self.del_beam_state,
                            doc.length, doc.c)
            for i in range(beam.width):
                state = <StateC*>beam.at(i)
                state.offset = offset
            offset += len(doc)
            beam.check_done(_beam_utils.check_final_state, NULL)
            beams.append(beam)
        return beams

    def get_oracle_sequence(self, doc, GoldParse gold):
        cdef Pool mem = Pool()
        # n_moves should not be zero at this point, but make sure to avoid zero-length mem alloc
        assert self.n_moves > 0
        costs = <float*>mem.alloc(self.n_moves, sizeof(float))
        is_valid = <int*>mem.alloc(self.n_moves, sizeof(int))

        cdef StateClass state = StateClass(doc, offset=0)
        self.initialize_state(state.c)
        history = []
        while not state.is_final():
            self.set_costs(is_valid, costs, state, gold)
            for i in range(self.n_moves):
                if is_valid[i] and costs[i] <= 0:
                    action = self.c[i]
                    history.append(i)
                    action.do(state.c, action.label)
                    break
            else:
                raise ValueError(Errors.E024)
        return history

    def apply_transition(self, StateClass state, name):
        if not self.is_valid(state, name):
            raise ValueError(Errors.E170.format(name=name))
        action = self.lookup_transition(name)
        action.do(state.c, action.label)

    cdef int initialize_state(self, StateC* state) nogil:
        pass

    cdef int finalize_state(self, StateC* state) nogil:
        pass

    def finalize_doc(self, doc):
        pass

    def preprocess_gold(self, GoldParse gold):
        raise NotImplementedError

    def is_gold_parse(self, StateClass state, GoldParse gold):
        raise NotImplementedError

    cdef Transition lookup_transition(self, object name) except *:
        raise NotImplementedError

    cdef Transition init_transition(self, int clas, int move, attr_t label) except *:
        raise NotImplementedError

    def is_valid(self, StateClass stcls, move_name):
        action = self.lookup_transition(move_name)
        if action.move == 0:
            return False
        return action.is_valid(stcls.c, action.label)

    cdef int set_valid(self, int* is_valid, const StateC* st) nogil:
        cdef int i
        for i in range(self.n_moves):
            is_valid[i] = self.c[i].is_valid(st, self.c[i].label)

    cdef int set_costs(self, int* is_valid, weight_t* costs,
                       StateClass stcls, GoldParse gold) except -1:
        cdef int i
        self.set_valid(is_valid, stcls.c)
        cdef int n_gold = 0
        for i in range(self.n_moves):
            if is_valid[i]:
                costs[i] = self.c[i].get_cost(stcls, &gold.c, self.c[i].label)
                n_gold += costs[i] <= 0
            else:
                costs[i] = 9000
        if n_gold <= 0:
            raise ValueError(Errors.E024)

    def get_class_name(self, int clas):
        act = self.c[clas]
        return self.move_name(act.move, act.label)

    def initialize_actions(self, labels_by_action, min_freq=None):
        self.labels = {}
        self.n_moves = 0
        added_labels = []
        added_actions = {}
        for action, label_freqs in sorted(labels_by_action.items()):
            action = int(action)
            # Make sure we take a copy here, and that we get a Counter
            self.labels[action] = Counter()
            # Have to be careful here: Sorting must be stable, or our model
            # won't be read back in correctly.
            sorted_labels = [(f, L) for L, f in label_freqs.items()]
            sorted_labels.sort()
            sorted_labels.reverse()
            for freq, label_str in sorted_labels:
                if freq < 0:
                    added_labels.append((freq, label_str))
                    added_actions.setdefault(label_str, []).append(action)
                else:
                    self.add_action(int(action), label_str)
                    self.labels[action][label_str] = freq
        added_labels.sort(reverse=True)
        for freq, label_str in added_labels:
            for action in added_actions[label_str]:
                self.add_action(int(action), label_str)
                self.labels[action][label_str] = freq

    def add_action(self, int action, label_name):
        cdef attr_t label_id
        if not isinstance(label_name, int) and \
           not isinstance(label_name, long):
            label_id = self.strings.add(label_name)
        else:
            label_id = label_name
        # Check we're not creating a move we already have, so that this is
        # idempotent
        for trans in self.c[:self.n_moves]:
            if trans.move == action and trans.label == label_id:
                return 0
        if self.n_moves >= self._size:
            self._size *= 2
            self.c = <Transition*>self.mem.realloc(self.c, self._size * sizeof(self.c[0]))
        self.c[self.n_moves] = self.init_transition(self.n_moves, action, label_id)
        self.n_moves += 1
        # Add the new (action, label) pair, making up a frequency for it if
        # necessary. To preserve sort order, the frequency needs to be lower
        # than previous frequencies.
        if self.labels.get(action, []):
            new_freq = min(self.labels[action].values())
        else:
            self.labels[action] = Counter()
            new_freq = -1
        if new_freq > 0:
            new_freq = 0
        self.labels[action][label_name] = new_freq-1
        return 1

    def to_disk(self, path, **kwargs):
        with path.open('wb') as file_:
            file_.write(self.to_bytes(**kwargs))

    def from_disk(self, path, **kwargs):
        with path.open('rb') as file_:
            byte_data = file_.read()
        self.from_bytes(byte_data, **kwargs)
        return self

    def to_bytes(self, exclude=tuple(), **kwargs):
        transitions = []
        serializers = {
            'moves': lambda: srsly.json_dumps(self.labels),
            'strings': lambda: self.strings.to_bytes()
        }
        exclude = util.get_serialization_exclude(serializers, exclude, kwargs)
        return util.to_bytes(serializers, exclude)

    def from_bytes(self, bytes_data, exclude=tuple(), **kwargs):
        labels = {}
        deserializers = {
            'moves': lambda b: labels.update(srsly.json_loads(b)),
            'strings': lambda b: self.strings.from_bytes(b)
        }
        exclude = util.get_serialization_exclude(deserializers, exclude, kwargs)
        msg = util.from_bytes(bytes_data, deserializers, exclude)
        self.initialize_actions(labels)
        return self



# coding: utf-8
from __future__ import unicode_literals

import pytest
from spacy.lang.en.lex_attrs import like_num


def test_en_tokenizer_handles_long_text(en_tokenizer):
    text = """Tributes pour in for late British Labour Party leader

Tributes poured in from around the world Thursday
to the late Labour Party leader John Smith, who died earlier from a massive
heart attack aged 55.

In Washington, the US State Department issued a statement regretting "the
untimely death" of the rapier-tongued Scottish barrister and parliamentarian.

"Mr. Smith, throughout his distinguished"""
    tokens = en_tokenizer(text)
    assert len(tokens) == 76


@pytest.mark.parametrize(
    "text,length",
    [
        ("The U.S. Army likes Shock and Awe.", 8),
        ("U.N. regulations are not a part of their concern.", 10),
        ("“Isn't it?”", 6),
        ("""Yes! "I'd rather have a walk", Ms. Comble sighed. """, 15),
        ("""'Me too!', Mr. P. Delaware cried. """, 11),
        ("They ran about 10km.", 6),
        pytest.param(
            "But then the 6,000-year ice age came...", 10, marks=pytest.mark.xfail()
        ),
    ],
)
def test_en_tokenizer_handles_cnts(en_tokenizer, text, length):
    tokens = en_tokenizer(text)
    assert len(tokens) == length


@pytest.mark.parametrize(
    "text,match",
    [
        ("10", True),
        ("1", True),
        ("10,000", True),
        ("10,00", True),
        ("999.0", True),
        ("one", True),
        ("two", True),
        ("billion", True),
        ("dog", False),
        (",", False),
        ("1/2", True),
    ],
)
def test_lex_attrs_like_number(en_tokenizer, text, match):
    tokens = en_tokenizer(text)
    assert len(tokens) == 1
    assert tokens[0].like_num == match


@pytest.mark.parametrize(
    "word",
    [
        "third",
        "Millionth",
        "100th",
        "Hundredth",
    ]
)
def test_en_lex_attrs_like_number_for_ordinal(word):
    assert like_num(word)


@pytest.mark.parametrize("word", ["eleven"])
def test_en_lex_attrs_capitals(word):
    assert like_num(word)
    assert like_num(word.upper())



# coding: utf-8
from __future__ import unicode_literals

from spacy.util import get_json_validator, validate_json, validate_schema
from spacy.cli._schemas import META_SCHEMA, TRAINING_SCHEMA
from spacy.matcher._schemas import TOKEN_PATTERN_SCHEMA
import pytest


@pytest.fixture(scope="session")
def training_schema_validator():
    return get_json_validator(TRAINING_SCHEMA)


def test_validate_schema():
    validate_schema({"type": "object"})
    with pytest.raises(Exception):
        validate_schema({"type": lambda x: x})


@pytest.mark.parametrize("schema", [TRAINING_SCHEMA, META_SCHEMA, TOKEN_PATTERN_SCHEMA])
def test_schemas(schema):
    validate_schema(schema)


@pytest.mark.parametrize(
    "data",
    [
        {"text": "Hello world"},
        {"text": "Hello", "ents": [{"start": 0, "end": 5, "label": "TEST"}]},
    ],
)
def test_json_schema_training_valid(data, training_schema_validator):
    errors = validate_json([data], training_schema_validator)
    assert not errors


@pytest.mark.parametrize(
    "data,n_errors",
    [
        ({"spans": []}, 1),
        ({"text": "Hello", "ents": [{"start": "0", "end": "5", "label": "TEST"}]}, 2),
        ({"text": "Hello", "ents": [{"start": 0, "end": 5}]}, 1),
        ({"text": "Hello", "ents": [{"start": 0, "end": 5, "label": "test"}]}, 1),
        ({"text": "spaCy", "tokens": [{"pos": "PROPN"}]}, 2),
    ],
)
def test_json_schema_training_invalid(data, n_errors, training_schema_validator):
    errors = validate_json([data], training_schema_validator)
    assert len(errors) == n_errors



# coding: utf8
from __future__ import unicode_literals

from .tokenizer_exceptions import TOKENIZER_EXCEPTIONS
from .tag_map import TAG_MAP
from .stop_words import STOP_WORDS
from .lex_attrs import LEX_ATTRS

from ...attrs import LANG
from ...language import Language
from ...tokens import Doc
from ...util import DummyTokenizer


class ThaiTokenizer(DummyTokenizer):
    def __init__(self, cls, nlp=None):
        try:
            from pythainlp.tokenize import word_tokenize
        except ImportError:
            raise ImportError(
                "The Thai tokenizer requires the PyThaiNLP library: "
                "https://github.com/PyThaiNLP/pythainlp"
            )

        self.word_tokenize = word_tokenize
        self.vocab = nlp.vocab if nlp is not None else cls.create_vocab(nlp)

    def __call__(self, text):
        words = list(self.word_tokenize(text))
        spaces = [False] * len(words)
        return Doc(self.vocab, words=words, spaces=spaces)


class ThaiDefaults(Language.Defaults):
    lex_attr_getters = dict(Language.Defaults.lex_attr_getters)
    lex_attr_getters.update(LEX_ATTRS)
    lex_attr_getters[LANG] = lambda _text: "th"
    tokenizer_exceptions = dict(TOKENIZER_EXCEPTIONS)
    tag_map = TAG_MAP
    stop_words = STOP_WORDS

    @classmethod
    def create_tokenizer(cls, nlp=None):
        return ThaiTokenizer(cls, nlp)


class Thai(Language):
    lang = "th"
    Defaults = ThaiDefaults

    def make_doc(self, text):
        return self.tokenizer(text)


__all__ = ["Thai"]



# coding: utf-8
from __future__ import unicode_literals

import pytest
from spacy.lookups import Lookups, Table
from spacy.strings import get_string_id
from spacy.vocab import Vocab

from ..util import make_tempdir


def test_lookups_api():
    table_name = "test"
    data = {"foo": "bar", "hello": "world"}
    lookups = Lookups()
    lookups.add_table(table_name, data)
    assert len(lookups) == 1
    assert table_name in lookups
    assert lookups.has_table(table_name)
    table = lookups.get_table(table_name)
    assert table.name == table_name
    assert len(table) == 2
    assert table["hello"] == "world"
    table["a"] = "b"
    assert table["a"] == "b"
    table = lookups.get_table(table_name)
    assert len(table) == 3
    with pytest.raises(KeyError):
        lookups.get_table("xyz")
    with pytest.raises(ValueError):
        lookups.add_table(table_name)
    table = lookups.remove_table(table_name)
    assert table.name == table_name
    assert len(lookups) == 0
    assert table_name not in lookups
    with pytest.raises(KeyError):
        lookups.get_table(table_name)


def test_table_api():
    table = Table(name="table")
    assert table.name == "table"
    assert len(table) == 0
    assert "abc" not in table
    data = {"foo": "bar", "hello": "world"}
    table = Table(name="table", data=data)
    assert len(table) == len(data)
    assert "foo" in table
    assert get_string_id("foo") in table
    assert table["foo"] == "bar"
    assert table[get_string_id("foo")] == "bar"
    assert table.get("foo") == "bar"
    assert table.get("abc") is None
    table["abc"] = 123
    assert table["abc"] == 123
    assert table[get_string_id("abc")] == 123
    table.set("def", 456)
    assert table["def"] == 456
    assert table[get_string_id("def")] == 456


def test_table_api_to_from_bytes():
    data = {"foo": "bar", "hello": "world", "abc": 123}
    table = Table(name="table", data=data)
    table_bytes = table.to_bytes()
    new_table = Table().from_bytes(table_bytes)
    assert new_table.name == "table"
    assert len(new_table) == 3
    assert new_table["foo"] == "bar"
    assert new_table[get_string_id("foo")] == "bar"
    new_table2 = Table(data={"def": 456})
    new_table2.from_bytes(table_bytes)
    assert len(new_table2) == 3
    assert "def" not in new_table2


@pytest.mark.skip(reason="This fails on Python 3.5")
def test_lookups_to_from_bytes():
    lookups = Lookups()
    lookups.add_table("table1", {"foo": "bar", "hello": "world"})
    lookups.add_table("table2", {"a": 1, "b": 2, "c": 3})
    lookups_bytes = lookups.to_bytes()
    new_lookups = Lookups()
    new_lookups.from_bytes(lookups_bytes)
    assert len(new_lookups) == 2
    assert "table1" in new_lookups
    assert "table2" in new_lookups
    table1 = new_lookups.get_table("table1")
    assert len(table1) == 2
    assert table1["foo"] == "bar"
    table2 = new_lookups.get_table("table2")
    assert len(table2) == 3
    assert table2["b"] == 2
    assert new_lookups.to_bytes() == lookups_bytes


@pytest.mark.skip(reason="This fails on Python 3.5")
def test_lookups_to_from_disk():
    lookups = Lookups()
    lookups.add_table("table1", {"foo": "bar", "hello": "world"})
    lookups.add_table("table2", {"a": 1, "b": 2, "c": 3})
    with make_tempdir() as tmpdir:
        lookups.to_disk(tmpdir)
        new_lookups = Lookups()
        new_lookups.from_disk(tmpdir)
    assert len(new_lookups) == 2
    assert "table1" in new_lookups
    assert "table2" in new_lookups
    table1 = new_lookups.get_table("table1")
    assert len(table1) == 2
    assert table1["foo"] == "bar"
    table2 = new_lookups.get_table("table2")
    assert len(table2) == 3
    assert table2["b"] == 2


@pytest.mark.skip(reason="This fails on Python 3.5")
def test_lookups_to_from_bytes_via_vocab():
    table_name = "test"
    vocab = Vocab()
    vocab.lookups.add_table(table_name, {"foo": "bar", "hello": "world"})
    assert table_name in vocab.lookups
    vocab_bytes = vocab.to_bytes()
    new_vocab = Vocab()
    new_vocab.from_bytes(vocab_bytes)
    assert len(new_vocab.lookups) == len(vocab.lookups)
    assert table_name in new_vocab.lookups
    table = new_vocab.lookups.get_table(table_name)
    assert len(table) == 2
    assert table["hello"] == "world"
    assert new_vocab.to_bytes() == vocab_bytes


@pytest.mark.skip(reason="This fails on Python 3.5")
def test_lookups_to_from_disk_via_vocab():
    table_name = "test"
    vocab = Vocab()
    vocab.lookups.add_table(table_name, {"foo": "bar", "hello": "world"})
    assert table_name in vocab.lookups
    with make_tempdir() as tmpdir:
        vocab.to_disk(tmpdir)
        new_vocab = Vocab()
        new_vocab.from_disk(tmpdir)
    assert len(new_vocab.lookups) == len(vocab.lookups)
    assert table_name in new_vocab.lookups
    table = new_vocab.lookups.get_table(table_name)
    assert len(table) == 2
    assert table["hello"] == "world"



# coding: utf8
from __future__ import unicode_literals

from ...symbols import POS, PUNCT, ADJ, CCONJ, SCONJ, NUM, DET, ADV, ADP, X
from ...symbols import NOUN, PROPN, PART, INTJ, SPACE, PRON, AUX, VERB


TAG_MAP = {
    "$(": {POS: PUNCT, "PunctType": "brck"},
    "$,": {POS: PUNCT, "PunctType": "comm"},
    "$.": {POS: PUNCT, "PunctType": "peri"},
    "ADJA": {POS: ADJ},
    "ADJD": {POS: ADJ},
    "ADV": {POS: ADV},
    "APPO": {POS: ADP, "AdpType": "post"},
    "APPR": {POS: ADP, "AdpType": "prep"},
    "APPRART": {POS: ADP, "AdpType": "prep", "PronType": "art"},
    "APZR": {POS: ADP, "AdpType": "circ"},
    "ART": {POS: DET, "PronType": "art"},
    "CARD": {POS: NUM, "NumType": "card"},
    "FM": {POS: X, "Foreign": "yes"},
    "ITJ": {POS: INTJ},
    "KOKOM": {POS: CCONJ, "ConjType": "comp"},
    "KON": {POS: CCONJ},
    "KOUI": {POS: SCONJ},
    "KOUS": {POS: SCONJ},
    "NE": {POS: PROPN},
    "NNE": {POS: PROPN},
    "NN": {POS: NOUN},
    "PAV": {POS: ADV, "PronType": "dem"},
    "PROAV": {POS: ADV, "PronType": "dem"},
    "PDAT": {POS: DET, "PronType": "dem"},
    "PDS": {POS: PRON, "PronType": "dem"},
    "PIAT": {POS: DET, "PronType": "ind|neg|tot"},
    "PIDAT": {POS: DET, "PronType": "ind|neg|tot"},
    "PIS": {POS: PRON, "PronType": "ind|neg|tot"},
    "PPER": {POS: PRON, "PronType": "prs"},
    "PPOSAT": {POS: DET, "Poss": "yes", "PronType": "prs"},
    "PPOSS": {POS: PRON, "Poss": "yes", "PronType": "prs"},
    "PRELAT": {POS: DET, "PronType": "rel"},
    "PRELS": {POS: PRON, "PronType": "rel"},
    "PRF": {POS: PRON, "PronType": "prs", "Reflex": "yes"},
    "PTKA": {POS: PART},
    "PTKANT": {POS: PART, "PartType": "res"},
    "PTKNEG": {POS: PART, "Polarity": "neg"},
    "PTKVZ": {POS: ADP, "PartType": "vbp"},
    "PTKZU": {POS: PART, "PartType": "inf"},
    "PWAT": {POS: DET, "PronType": "int"},
    "PWAV": {POS: ADV, "PronType": "int"},
    "PWS": {POS: PRON, "PronType": "int"},
    "TRUNC": {POS: X, "Hyph": "yes"},
    "VAFIN": {POS: AUX, "Mood": "ind", "VerbForm": "fin"},
    "VAIMP": {POS: AUX, "Mood": "imp", "VerbForm": "fin"},
    "VAINF": {POS: AUX, "VerbForm": "inf"},
    "VAPP": {POS: AUX, "Aspect": "perf", "VerbForm": "part"},
    "VMFIN": {POS: VERB, "Mood": "ind", "VerbForm": "fin", "VerbType": "mod"},
    "VMINF": {POS: VERB, "VerbForm": "inf", "VerbType": "mod"},
    "VMPP": {POS: VERB, "Aspect": "perf", "VerbForm": "part", "VerbType": "mod"},
    "VVFIN": {POS: VERB, "Mood": "ind", "VerbForm": "fin"},
    "VVIMP": {POS: VERB, "Mood": "imp", "VerbForm": "fin"},
    "VVINF": {POS: VERB, "VerbForm": "inf"},
    "VVIZU": {POS: VERB, "VerbForm": "inf"},
    "VVPP": {POS: VERB, "Aspect": "perf", "VerbForm": "part"},
    "XY": {POS: X},
    "_SP": {POS: SPACE},
}



# coding: utf-8
from __future__ import unicode_literals

import pytest
from spacy.util import get_lang_class


# fmt: off
# Only include languages with no external dependencies
# excluded: ja, ru, th, uk, vi, zh
LANGUAGES = ["af", "ar", "bg", "bn", "ca", "cs", "da", "de", "el", "en", "es",
             "et", "fa", "fi", "fr", "ga", "he", "hi", "hr", "hu", "id", "is",
             "it", "kn", "lt", "lv", "nb", "nl", "pl", "pt", "ro", "si", "sk",
             "sl", "sq", "sr", "sv", "ta", "te", "tl", "tr", "tt", "ur", 'yo']
# fmt: on


@pytest.mark.parametrize("lang", LANGUAGES)
def test_lang_initialize(lang, capfd):
    """Test that languages can be initialized."""
    nlp = get_lang_class(lang)()
    # Check for stray print statements (see #3342)
    doc = nlp("test")  # noqa: F841
    captured = capfd.readouterr()
    assert not captured.out



# coding: utf8
from __future__ import unicode_literals

from spacy.lang.en import English
from spacy.tokens import Doc
from spacy.vocab import Vocab


def test_issue4133(en_vocab):
    nlp = English()
    vocab_bytes = nlp.vocab.to_bytes()
    words = ["Apple", "is", "looking", "at", "buying", "a", "startup"]
    pos = ["NOUN", "VERB", "ADP", "VERB", "PROPN", "NOUN", "ADP"]
    doc = Doc(en_vocab, words=words)
    for i, token in enumerate(doc):
        token.pos_ = pos[i]

    # usually this is already True when starting from proper models instead of blank English
    doc.is_tagged = True

    doc_bytes = doc.to_bytes()

    vocab = Vocab()
    vocab = vocab.from_bytes(vocab_bytes)
    doc = Doc(vocab).from_bytes(doc_bytes)

    actual = []
    for token in doc:
        actual.append(token.pos_)

    assert actual == pos



# coding: utf8
from __future__ import unicode_literals

from spacy.matcher import Matcher, PhraseMatcher
from spacy.vocab import Vocab


def test_issue4373():
    """Test that PhraseMatcher.vocab can be accessed (like Matcher.vocab)."""
    matcher = Matcher(Vocab())
    assert isinstance(matcher.vocab, Vocab)
    matcher = PhraseMatcher(Vocab())
    assert isinstance(matcher.vocab, Vocab)



# coding: utf8
from __future__ import unicode_literals

from thinc.t2v import Pooling, max_pool, mean_pool
from thinc.neural._classes.difference import Siamese, CauchySimilarity

from .pipes import Pipe
from ..language import component
from .._ml import link_vectors_to_models


@component("sentencizer_hook", assigns=["doc.user_hooks"])
class SentenceSegmenter(object):
    """A simple spaCy hook, to allow custom sentence boundary detection logic
    (that doesn't require the dependency parse). To change the sentence
    boundary detection strategy, pass a generator function `strategy` on
    initialization, or assign a new strategy to the .strategy attribute.
    Sentence detection strategies should be generators that take `Doc` objects
    and yield `Span` objects for each sentence.
    """

    def __init__(self, vocab, strategy=None):
        self.vocab = vocab
        if strategy is None or strategy == "on_punct":
            strategy = self.split_on_punct
        self.strategy = strategy

    def __call__(self, doc):
        doc.user_hooks["sents"] = self.strategy
        return doc

    @staticmethod
    def split_on_punct(doc):
        start = 0
        seen_period = False
        for i, token in enumerate(doc):
            if seen_period and not token.is_punct:
                yield doc[start : token.i]
                start = token.i
                seen_period = False
            elif token.text in [".", "!", "?"]:
                seen_period = True
        if start < len(doc):
            yield doc[start : len(doc)]


@component("similarity", assigns=["doc.user_hooks"])
class SimilarityHook(Pipe):
    """
    Experimental: A pipeline component to install a hook for supervised
    similarity into `Doc` objects. Requires a `Tensorizer` to pre-process
    documents. The similarity model can be any object obeying the Thinc `Model`
    interface. By default, the model concatenates the elementwise mean and
    elementwise max of the two tensors, and compares them using the
    Cauchy-like similarity function from Chen (2013):

        >>> similarity = 1. / (1. + (W * (vec1-vec2)**2).sum())

    Where W is a vector of dimension weights, initialized to 1.
    """

    def __init__(self, vocab, model=True, **cfg):
        self.vocab = vocab
        self.model = model
        self.cfg = dict(cfg)

    @classmethod
    def Model(cls, length):
        return Siamese(Pooling(max_pool, mean_pool), CauchySimilarity(length))

    def __call__(self, doc):
        """Install similarity hook"""
        doc.user_hooks["similarity"] = self.predict
        return doc

    def pipe(self, docs, **kwargs):
        for doc in docs:
            yield self(doc)

    def predict(self, doc1, doc2):
        self.require_model()
        return self.model.predict([(doc1, doc2)])

    def update(self, doc1_doc2, golds, sgd=None, drop=0.0):
        self.require_model()
        sims, bp_sims = self.model.begin_update(doc1_doc2, drop=drop)

    def begin_training(self, _=tuple(), pipeline=None, sgd=None, **kwargs):
        """Allocate model, using width from tensorizer in pipeline.

        gold_tuples (iterable): Gold-standard training data.
        pipeline (list): The pipeline the model is part of.
        """
        if self.model is True:
            self.model = self.Model(pipeline[0].model.nO)
            link_vectors_to_models(self.vocab)
        if sgd is None:
            sgd = self.create_optimizer()
        return sgd



# encoding: utf8
from __future__ import unicode_literals


STOP_WORDS = set(
    """
a ach ag agus an aon ar arna as

ba beirt bhúr

caoga ceathair ceathrar chomh chuig chun cois céad cúig cúigear

daichead dar de deich deichniúr den dhá do don dtí dá dár dó

faoi faoin faoina faoinár fara fiche

gach gan go gur

haon hocht

i iad idir in ina ins inár is

le leis lena lenár

mar mo muid mé

na nach naoi naonúr ná ní níor nó nócha

ocht ochtar ochtó os

roimh

sa seacht seachtar seachtó seasca seisear siad sibh sinn sna sé sí

tar thar thú triúr trí trína trínár tríocha tú

um

ár

é éis

í

ó ón óna ónár
""".split()
)



# coding: utf8
from __future__ import unicode_literals

from ...attrs import LIKE_NUM


_num_words = [
    "cero",
    "uno",
    "dos",
    "tres",
    "cuatro",
    "cinco",
    "seis",
    "siete",
    "ocho",
    "nueve",
    "diez",
    "once",
    "doce",
    "trece",
    "catorce",
    "quince",
    "dieciséis",
    "diecisiete",
    "dieciocho",
    "diecinueve",
    "veinte",
    "veintiuno",
    "veintidós",
    "veintitrés",
    "veinticuatro",
    "veinticinco",
    "veintiséis",
    "veintisiete",
    "veintiocho",
    "veintinueve",
    "treinta",
    "cuarenta",
    "cincuenta",
    "sesenta",
    "setenta",
    "ochenta",
    "noventa",
    "cien",
    "mil",
    "millón",
    "billón",
    "trillón",
]


def like_num(text):
    if text.startswith(("+", "-", "±", "~")):
        text = text[1:]
    text = text.replace(",", "").replace(".", "")
    if text.isdigit():
        return True
    if text.count("/") == 1:
        num, denom = text.split("/")
        if num.isdigit() and denom.isdigit():
            return True
    if text.lower() in _num_words:
        return True
    return False


LEX_ATTRS = {LIKE_NUM: like_num}



# coding: utf8
from __future__ import unicode_literals

from .stop_words import STOP_WORDS
from .lex_attrs import LEX_ATTRS
from .punctuation import TOKENIZER_SUFFIXES

from ..tokenizer_exceptions import BASE_EXCEPTIONS
from ...language import Language
from ...attrs import LANG


class BasqueDefaults(Language.Defaults):
    lex_attr_getters = dict(Language.Defaults.lex_attr_getters)
    lex_attr_getters.update(LEX_ATTRS)
    lex_attr_getters[LANG] = lambda text: "eu"

    tokenizer_exceptions = BASE_EXCEPTIONS
    stop_words = STOP_WORDS
    suffixes = TOKENIZER_SUFFIXES


class Basque(Language):
    lang = "eu"
    Defaults = BasqueDefaults


__all__ = ["Basque"]



# coding: utf8
from __future__ import unicode_literals

from ...symbols import ORTH, LEMMA, TAG, NORM, PRON_LEMMA


_exc = {}


_abbr_exc = [
    {ORTH: "м", LEMMA: "метар", NORM: "метар"},
    {ORTH: "мм", LEMMA: "милиметар", NORM: "милиметар"},
    {ORTH: "цм", LEMMA: "центиметар", NORM: "центиметар"},
    {ORTH: "см", LEMMA: "сантиметар", NORM: "сантиметар"},
    {ORTH: "дм", LEMMA: "дециметар", NORM: "дециметар"},
    {ORTH: "км", LEMMA: "километар", NORM: "километар"},
    {ORTH: "кг", LEMMA: "килограм", NORM: "килограм"},
    {ORTH: "дкг", LEMMA: "декаграм", NORM: "декаграм"},
    {ORTH: "дг", LEMMA: "дециграм", NORM: "дециграм"},
    {ORTH: "мг", LEMMA: "милиграм", NORM: "милиграм"},
    {ORTH: "г", LEMMA: "грам", NORM: "грам"},
    {ORTH: "т", LEMMA: "тон", NORM: "тон"},
    {ORTH: "кл", LEMMA: "килолитар", NORM: "килолитар"},
    {ORTH: "хл", LEMMA: "хектолитар", NORM: "хектолитар"},
    {ORTH: "дкл", LEMMA: "декалитар", NORM: "декалитар"},
    {ORTH: "л", LEMMA: "литар", NORM: "литар"},
    {ORTH: "дл", LEMMA: "децилитар", NORM: "децилитар"}

]
for abbr in _abbr_exc:
    _exc[abbr[ORTH]] = [abbr]

_abbr_line_exc = [
    {ORTH: "д-р", LEMMA: "доктор", NORM: "доктор"},
    {ORTH: "м-р", LEMMA: "магистер", NORM: "магистер"},
    {ORTH: "г-ѓа", LEMMA: "госпоѓа", NORM: "госпоѓа"},
    {ORTH: "г-ца", LEMMA: "госпоѓица", NORM: "госпоѓица"},
    {ORTH: "г-дин", LEMMA: "господин", NORM: "господин"},

]

for abbr in _abbr_line_exc:
    _exc[abbr[ORTH]] = [abbr]

_abbr_dot_exc = [
    {ORTH: "в.", LEMMA: "век", NORM: "век"},
    {ORTH: "в.д.", LEMMA: "вршител на должност", NORM: "вршител на должност"},
    {ORTH: "г.", LEMMA: "година", NORM: "година"},
    {ORTH: "г.г.", LEMMA: "господин господин", NORM: "господин господин"},
    {ORTH: "м.р.", LEMMA: "машки род", NORM: "машки род"},
    {ORTH: "год.", LEMMA: "женски род", NORM: "женски род"},
    {ORTH: "с.р.", LEMMA: "среден род", NORM: "среден род"},
    {ORTH: "н.е.", LEMMA: "наша ера", NORM: "наша ера"},
    {ORTH: "о.г.", LEMMA: "оваа година", NORM: "оваа година"},
    {ORTH: "о.м.", LEMMA: "овој месец", NORM: "овој месец"},
    {ORTH: "с.", LEMMA: "село", NORM: "село"},
    {ORTH: "т.", LEMMA: "точка", NORM: "точка"},
    {ORTH: "т.е.", LEMMA: "то ест", NORM: "то ест"},
    {ORTH: "т.н.", LEMMA: "таканаречен", NORM: "таканаречен"},

    {ORTH: "бр.", LEMMA: "број", NORM: "број"},
    {ORTH: "гр.", LEMMA: "град", NORM: "град"},
    {ORTH: "др.", LEMMA: "другар", NORM: "другар"},
    {ORTH: "и др.", LEMMA: "и друго", NORM: "и друго"},
    {ORTH: "и сл.", LEMMA: "и слично", NORM: "и слично"},
    {ORTH: "кн.", LEMMA: "книга", NORM: "книга"},
    {ORTH: "мн.", LEMMA: "множина", NORM: "множина"},
    {ORTH: "на пр.", LEMMA: "на пример", NORM: "на пример"},
    {ORTH: "св.", LEMMA: "свети", NORM: "свети"},
    {ORTH: "сп.", LEMMA: "списание", NORM: "списание"},
    {ORTH: "с.", LEMMA: "страница", NORM: "страница"},
    {ORTH: "стр.", LEMMA: "страница", NORM: "страница"},
    {ORTH: "чл.", LEMMA: "член", NORM: "член"},

    {ORTH: "арх.", LEMMA: "архитект", NORM: "архитект"},
    {ORTH: "бел.", LEMMA: "белешка", NORM: "белешка"},
    {ORTH: "гимн.", LEMMA: "гимназија", NORM: "гимназија"},
    {ORTH: "ден.", LEMMA: "денар", NORM: "денар"},
    {ORTH: "ул.", LEMMA: "улица", NORM: "улица"},
    {ORTH: "инж.", LEMMA: "инженер", NORM: "инженер"},
    {ORTH: "проф.", LEMMA: "професор", NORM: "професор"},
    {ORTH: "студ.", LEMMA: "студент", NORM: "студент"},
    {ORTH: "бот.", LEMMA: "ботаника", NORM: "ботаника"},
    {ORTH: "мат.", LEMMA: "математика", NORM: "математика"},
    {ORTH: "мед.", LEMMA: "медицина", NORM: "медицина"},
    {ORTH: "прил.", LEMMA: "прилог", NORM: "прилог"},
    {ORTH: "прид.", LEMMA: "придавка", NORM: "придавка"},
    {ORTH: "сврз.", LEMMA: "сврзник", NORM: "сврзник"},
    {ORTH: "физ.", LEMMA: "физика", NORM: "физика"},
    {ORTH: "хем.", LEMMA: "хемија", NORM: "хемија"},
    {ORTH: "пр. н.", LEMMA: "природни науки", NORM: "природни науки"},
    {ORTH: "истор.", LEMMA: "историја", NORM: "историја"},
    {ORTH: "геогр.", LEMMA: "географија", NORM: "географија"},
    {ORTH: "литер.", LEMMA: "литература", NORM: "литература"},


]

for abbr in _abbr_dot_exc:
    _exc[abbr[ORTH]] = [abbr]


TOKENIZER_EXCEPTIONS = _exc



# coding: utf8
from __future__ import unicode_literals

from collections import OrderedDict

from ...lemmatizer import Lemmatizer
from ...parts_of_speech import NAMES as UPOS_NAMES


class MacedonianLemmatizer(Lemmatizer):
    def __call__(self, string, univ_pos, morphology=None):
        lookup_table = self.lookups.get_table("lemma_lookup", {})
        if "lemma_rules" not in self.lookups:
            return [lookup_table.get(string, string)]
        if isinstance(univ_pos, int):
            univ_pos = UPOS_NAMES.get(univ_pos, "X")
        univ_pos = univ_pos.lower()

        if univ_pos in ("", "eol", "space"):
            return [string.lower()]

        if string[-3:] == 'јќи':
            string = string[:-3]
            univ_pos = "verb"

        if callable(self.is_base_form) and self.is_base_form(univ_pos, morphology):
            return [string.lower()]
        index_table = self.lookups.get_table("lemma_index", {})
        exc_table = self.lookups.get_table("lemma_exc", {})
        rules_table = self.lookups.get_table("lemma_rules", {})
        if not any((index_table.get(univ_pos), exc_table.get(univ_pos), rules_table.get(univ_pos))):
            if univ_pos == "propn":
                return [string]
            else:
                return [string.lower()]
        lemmas = self.lemmatize(
            string,
            index_table.get(univ_pos, {}),
            exc_table.get(univ_pos, {}),
            rules_table.get(univ_pos, []),
        )
        return lemmas

    def lemmatize(self, string, index, exceptions, rules):
        orig = string
        string = string.lower()
        forms = []

        for old, new in rules:
            if string.endswith(old):
                form = string[: len(string) - len(old)] + new
                if not form:
                    continue
                if form in index or not form.isalpha():
                    forms.append(form)

        forms = list(OrderedDict.fromkeys(forms))
        for form in exceptions.get(string, []):
            if form not in forms:
                forms.insert(0, form)
        if not forms:
            forms.append(orig)

        return forms



# coding: utf8
from __future__ import unicode_literals

from ..punctuation import TOKENIZER_PREFIXES as BASE_TOKENIZER_PREFIXES
from ..punctuation import TOKENIZER_SUFFIXES as BASE_TOKENIZER_SUFFIXES
from ..punctuation import TOKENIZER_INFIXES as BASE_TOKENIZER_INFIXES

_prefixes = [r"\w{1,3}\$"] + BASE_TOKENIZER_PREFIXES

_suffixes = BASE_TOKENIZER_SUFFIXES

_infixes = [r"(\w+-\w+(-\w+)*)"] + BASE_TOKENIZER_INFIXES

TOKENIZER_PREFIXES = _prefixes
TOKENIZER_SUFFIXES = _suffixes
TOKENIZER_INFIXES = _infixes



# coding: utf8
from __future__ import unicode_literals

from spacy.lang.en import English
import pytest


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_issue3880():
    """Test that `nlp.pipe()` works when an empty string ends the batch.

    Fixed in v7.0.5 of Thinc.
    """
    texts = ["hello", "world", "", ""]
    nlp = English()
    nlp.add_pipe(nlp.create_pipe("parser"))
    nlp.add_pipe(nlp.create_pipe("ner"))
    nlp.add_pipe(nlp.create_pipe("tagger"))
    nlp.get_pipe("parser").add_label("dep")
    nlp.get_pipe("ner").add_label("PERSON")
    nlp.get_pipe("tagger").add_label("NN")
    nlp.begin_training()
    for doc in nlp.pipe(texts):
        pass



# coding: utf8
from __future__ import unicode_literals, division, print_function

import plac
from timeit import default_timer as timer
from wasabi import msg

from ..gold import GoldCorpus
from .. import util
from .. import displacy


@plac.annotations(
    model=("Model name or path", "positional", None, str),
    data_path=("Location of JSON-formatted evaluation data", "positional", None, str),
    gold_preproc=("Use gold preprocessing", "flag", "G", bool),
    gpu_id=("Use GPU", "option", "g", int),
    displacy_path=("Directory to output rendered parses as HTML", "option", "dp", str),
    displacy_limit=("Limit of parses to render as HTML", "option", "dl", int),
    return_scores=("Return dict containing model scores", "flag", "R", bool),
)
def evaluate(
    model,
    data_path,
    gpu_id=-1,
    gold_preproc=False,
    displacy_path=None,
    displacy_limit=25,
    return_scores=False,
):
    """
    Evaluate a model. To render a sample of parses in a HTML file, set an
    output directory as the displacy_path argument.
    """
    util.fix_random_seed()
    if gpu_id >= 0:
        util.use_gpu(gpu_id)
    util.set_env_log(False)
    data_path = util.ensure_path(data_path)
    displacy_path = util.ensure_path(displacy_path)
    if not data_path.exists():
        msg.fail("Evaluation data not found", data_path, exits=1)
    if displacy_path and not displacy_path.exists():
        msg.fail("Visualization output directory not found", displacy_path, exits=1)
    corpus = GoldCorpus(data_path, data_path)
    if model.startswith("blank:"):
        nlp = util.get_lang_class(model.replace("blank:", ""))()
    else:
        nlp = util.load_model(model)
    dev_docs = list(corpus.dev_docs(nlp, gold_preproc=gold_preproc))
    begin = timer()
    scorer = nlp.evaluate(dev_docs, verbose=False)
    end = timer()
    nwords = sum(len(doc_gold[0]) for doc_gold in dev_docs)
    results = {
        "Time": "%.2f s" % (end - begin),
        "Words": nwords,
        "Words/s": "%.0f" % (nwords / (end - begin)),
        "TOK": "%.2f" % scorer.token_acc,
        "POS": "%.2f" % scorer.tags_acc,
        "UAS": "%.2f" % scorer.uas,
        "LAS": "%.2f" % scorer.las,
        "NER P": "%.2f" % scorer.ents_p,
        "NER R": "%.2f" % scorer.ents_r,
        "NER F": "%.2f" % scorer.ents_f,
        "Textcat": "%.2f" % scorer.textcat_score,
    }
    msg.table(results, title="Results")

    if displacy_path:
        docs, golds = zip(*dev_docs)
        render_deps = "parser" in nlp.meta.get("pipeline", [])
        render_ents = "ner" in nlp.meta.get("pipeline", [])
        render_parses(
            docs,
            displacy_path,
            model_name=model,
            limit=displacy_limit,
            deps=render_deps,
            ents=render_ents,
        )
        msg.good("Generated {} parses as HTML".format(displacy_limit), displacy_path)
    if return_scores:
        return scorer.scores


def render_parses(docs, output_path, model_name="", limit=250, deps=True, ents=True):
    docs[0].user_data["title"] = model_name
    if ents:
        html = displacy.render(docs[:limit], style="ent", page=True)
        with (output_path / "entities.html").open("w", encoding="utf8") as file_:
            file_.write(html)
    if deps:
        html = displacy.render(
            docs[:limit], style="dep", page=True, options={"compact": True}
        )
        with (output_path / "parses.html").open("w", encoding="utf8") as file_:
            file_.write(html)



# coding: utf8
from __future__ import unicode_literals

from ...attrs import LIKE_NUM

_num_words = [
    "nula",
    "jeden",
    "dva",
    "tri",
    "štyri",
    "päť",
    "šesť",
    "sedem",
    "osem",
    "deväť",
    "desať",
    "jedenásť",
    "dvanásť",
    "trinásť",
    "štrnásť",
    "pätnásť",
    "šestnásť",
    "sedemnásť",
    "osemnásť",
    "devätnásť",
    "dvadsať",
    "tridsať",
    "štyridsať",
    "päťdesiat",
    "šesťdesiat",
    "sedemdesiat",
    "osemdesiat",
    "deväťdesiat",
    "sto",
    "tisíc",
    "milión",
    "miliarda",
    "bilión",
    "biliarda",
    "trilión",
    "triliarda",
    "kvadrilión",
]


def like_num(text):
    if text.startswith(("+", "-", "±", "~")):
        text = text[1:]
    text = text.replace(",", "").replace(".", "")
    if text.isdigit():
        return True
    if text.count("/") == 1:
        num, denom = text.split("/")
        if num.isdigit() and denom.isdigit():
            return True
    if text.lower() in _num_words:
        return True
    return False


LEX_ATTRS = {LIKE_NUM: like_num}



# coding: utf8
from __future__ import unicode_literals

from spacy.lang.en import English
from spacy.util import fix_random_seed


def test_issue6177():
    """Test that after fixing the random seed, the results of the pipeline are truly identical"""

    # NOTE: no need to transform this code to v3 when 'master' is merged into 'develop'.
    # A similar test exists already for v3: test_issue5551
    # This is just a backport
    results = []
    for i in range(3):
        fix_random_seed(0)
        nlp = English()
        example = (
            "Once hot, form ping-pong-ball-sized balls of the mixture, each weighing roughly 25 g.",
            {"cats": {"Labe1": 1.0, "Label2": 0.0, "Label3": 0.0}},
        )
        textcat = nlp.create_pipe("textcat")
        nlp.add_pipe(textcat)
        for label in set(example[1]["cats"]):
            textcat.add_label(label)
        # Train
        optimizer = nlp.begin_training()
        text, annots = example
        nlp.update([text], [annots], sgd=optimizer)
        # Store the result of each iteration
        result = textcat.model.predict([nlp.make_doc(text)])
        results.append(list(result[0]))

    # All results should be the same because of the fixed seed
    assert len(results) == 3
    assert results[0] == results[1]
    assert results[0] == results[2]



# coding: utf8
from __future__ import unicode_literals

from ...attrs import LIKE_NUM

_num_words = [
    "אפס",
    "אחד",
    "אחת",
    "שתיים",
    "שתים",
    "שניים",
    "שנים",
    "שלוש",
    "שלושה",
    "ארבע",
    "ארבעה",
    "חמש",
    "חמישה",
    "שש",
    "שישה",
    "שבע",
    "שבעה",
    "שמונה",
    "תשע",
    "תשעה",
    "עשר",
    "עשרה",
    "אחד עשר",
    "אחת עשרה",
    "שנים עשר",
    "שתים עשרה",
    "שלושה עשר",
    "שלוש עשרה",
    "ארבעה עשר",
    "ארבע עשרה",
    "חמישה עשר",
    "חמש עשרה",
    "ששה עשר",
    "שש עשרה",
    "שבעה עשר",
    "שבע עשרה",
    "שמונה עשר",
    "שמונה עשרה",
    "תשעה עשר",
    "תשע עשרה",
    "עשרים",
    "שלושים",
    "ארבעים",
    "חמישים",
    "שישים",
    "שבעים",
    "שמונים",
    "תשעים",
    "מאה",
    "אלף",
    "מליון",
    "מליארד",
    "טריליון",
]


_ordinal_words = [
    "ראשון",
    "שני",
    "שלישי",
    "רביעי",
    "חמישי",
    "שישי",
    "שביעי",
    "שמיני",
    "תשיעי",
    "עשירי",
]

def like_num(text):
    if text.startswith(("+", "-", "±", "~")):
        text = text[1:]
    text = text.replace(",", "").replace(".", "")
    if text.isdigit():
        return True

    if text.count("/") == 1:
        num, denom = text.split("/")
        if num.isdigit() and denom.isdigit():
            return True
    
    if text in _num_words:
        return True

    # CHeck ordinal number
    if text in _ordinal_words:
        return True
    return False


LEX_ATTRS = {LIKE_NUM: like_num}
